# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves the main function.
"""

import argparse
import sys
import time

from msprobe.msaccucmp.cmp_utils import log, file_utils
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse.dump_data_parser import DumpDataParser


def _save_log_parser(save_log_parser: argparse.ArgumentParser) -> None:
    save_log_parser.add_argument(
        '-d', '--dump_file', dest='dump_path', default='', type=safe_path_string,
        help='<Required> the dump file path, supports one AICPU custom operator dump file.',
        required=True)
    save_log_parser.add_argument('-out', '--output', dest='output_path', default='', type=safe_path_string,
                                 help='<Optional> the output path')


def _do_cmd() -> int:
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='commands')
    save_log_parser = subparsers.add_parser('save_log', help='Save AICPU custom operator log.')

    _save_log_parser(save_log_parser)

    args = parser.parse_args(sys.argv[1:])
    if len(sys.argv) < 2:
        parser.print_help()
        raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    if sys.argv[1] == 'save_log':
        args.dump_version = 2
        args.output_file_type = None
        ret = _do_save_log(args)
        return ret
    else:
        return 0


def _do_save_log(args: argparse.Namespace) -> int:
    ret = DumpDataParser(args).parse_log_data()
    return ret


def main() -> None:
    """
    parse argument and run command
    :return:
    """
    start = time.time()
    with file_utils.UmaskWrapper():
        try:
            ret = _do_cmd()
        except CompareError as err:
            ret = err.code
        except Exception as base_err:
            log.print_error_log(f'Basic error running {sys.argv[0]}: {base_err}')
            sys.exit(1)
    end = time.time()
    if ret != 0:
        log.print_error_log("Failed to parse dump log.")
    log.print_info_log(
        'The command was completed and took %d seconds.' % (end - start))


if __name__ == '__main__':
    main()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import os
from typing import Union

from msprobe.infer.utils.check.checker import Checker, CheckResult, rule


WHITE_LIST_PATTERN = r"^(?!.*\.\.)(?!:)[a-zA-Z0-9_./:-]+$"
BLACK_LIST_PATTERN = r"[^_A-Za-z0-9/.-]"
IDS_PATTERN = r'^(\d+(?:_\d+)*)(,\d+(?:_\d+)*)*$'
INVALID_CHAR = "[\n\f\r\b\t\v\u000D\u000A\u000C\u000B\u0009\u0008\u007F&%$*^#@;]"


class StringChecker(Checker):

    @rule()
    def is_str(self) -> Union["StringChecker", CheckResult]:
        err_msg = f"{self.instance} is not a string"
        return isinstance(self.instance, str), err_msg

    @rule()
    def is_file_name_too_long(self) -> Union["StringChecker", CheckResult]:
        err_msg = "File name too long"
        ret = len(self.instance) > 4095 or any(map(lambda s: len(s) > 255, self.instance.split(os.path.sep)))
        return not ret, err_msg

    @rule()
    def is_str_safe(self) -> Union["StringChecker", CheckResult]:
        err_msg = "String parameter contains invalid characters"
        return re.search(WHITE_LIST_PATTERN, self.instance), err_msg
    
    @rule()
    def is_str_valid_bool(self) -> Union["StringChecker", CheckResult]:
        err_msg = "Boolean value expected 'yes', 'y', 'Y', 'YES', 'true', 't', 'TRUE', 'True', '1' for true"
        return self.instance.lower() in ('yes', 'y', 'Y', 'YES', 'true', 't', 'TRUE', 'True', '1'), err_msg
    
    @rule()
    def is_str_valid_path(self) -> Union["StringChecker", CheckResult]:
        err_msg = "Input path contains invalid characters"
        return not re.search(BLACK_LIST_PATTERN, self.instance), err_msg
    
    @rule()
    def is_str_valid_ids(self):
        err_msg = f"dym range string \"{self.instance}\" is not a legal string"
        return re.match(IDS_PATTERN, self.instance), err_msg

    @rule()
    def str_has_no_invalid_char(self):
        err_msg = "Input string contains invalid chars" 
        return not re.search(INVALID_CHAR, self.instance), err_msg



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from .hierarchy import Hierarchy
from ..utils.global_state import GraphState
from ..utils.constant import NPU, BENCH, SINGLE


class LayoutHierarchyModel:
    npu_hierarchy = None
    bench_hierarchy = None
    single_hierarchy = None
      
    hierarchy = {
        NPU: npu_hierarchy,
        BENCH: bench_hierarchy,
        SINGLE: single_hierarchy
    }    

    @staticmethod
    def change_expand_state(node_name, graph_type, repo, micro_step, rank_step):
        if node_name == 'root':
            rank = rank_step.get('rank')
            step = rank_step.get('step')
            GraphState.set_global_value("update_precision_cache", {})  # 切换图清缓存
            LayoutHierarchyModel.hierarchy[graph_type] = Hierarchy(graph_type, repo, micro_step, rank, step)
            
        elif LayoutHierarchyModel.hierarchy.get(graph_type, None):
            LayoutHierarchyModel.hierarchy[graph_type].update_graph_data(node_name)
            LayoutHierarchyModel.hierarchy[graph_type].update_graph_shape()
            LayoutHierarchyModel.hierarchy[graph_type].update_graph_position()
        else:
            return {}
        return LayoutHierarchyModel.hierarchy[graph_type].get_hierarchy()

    @staticmethod
    def update_hierarchy_data(graph_type):
        if LayoutHierarchyModel.hierarchy.get(graph_type, None):
            return LayoutHierarchyModel.hierarchy[graph_type].update_hierarchy_data()
        else:
            return {}
        
    @staticmethod
    def update_current_hierarchy_data(data):
        npu_update_data = []
        bench_update_data = []
        for node in data:
            if node['graph_type'] == NPU:
                npu_update_data.append(node)
            elif node['graph_type'] == BENCH:
                bench_update_data.append(node)
        if LayoutHierarchyModel.hierarchy.get(NPU, None) and npu_update_data:
            LayoutHierarchyModel.hierarchy[NPU].update_current_hierarchy_data(npu_update_data)
        if LayoutHierarchyModel.hierarchy.get(BENCH, None) and bench_update_data:
            LayoutHierarchyModel.hierarchy[BENCH].update_current_hierarchy_data(bench_update_data)
     



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from tqdm import tqdm
from dataclasses import dataclass
from functools import wraps
from typing import Any, Dict, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Rectangle
from msprobe.core.common.utils import is_int
from msprobe.core.common.file_utils import create_directory, remove_path
from msprobe.core.common.log import logger
from msprobe.core.common.parallel_state import RankGroupGenerator


class MegatronStepInfo:
    is_megatron = False
    is_forward = False
    is_backward = False
    forward_micro_step = -1
    backward_micro_step = -1

    @classmethod
    def reset(cls):
        """重置所有类属性到初始状态"""
        cls.is_megatron = False
        cls.is_forward = False
        cls.is_backward = False
        cls.forward_micro_step = -1
        cls.backward_micro_step = -1


def wrap_megatron_step(func, is_forward=True):
    @wraps(func)
    def wrapped_func(*args, **kwargs):
        if not MegatronStepInfo.is_megatron:
            MegatronStepInfo.is_megatron = True
        if is_forward:
            MegatronStepInfo.is_forward = True
            MegatronStepInfo.is_backward = False
            MegatronStepInfo.forward_micro_step += 1
        else:
            MegatronStepInfo.is_forward = False
            MegatronStepInfo.is_backward = True
            MegatronStepInfo.backward_micro_step += 1
        return func(*args, **kwargs)

    return wrapped_func


def get_micro_step():
    return MegatronStepInfo.forward_micro_step if MegatronStepInfo.is_forward else MegatronStepInfo.backward_micro_step


def is_megatron():
    return MegatronStepInfo.is_megatron


@dataclass
class VisualizationConfig:
    """可视化配置数据类"""
    height_per_tp: float = 0.5
    width_per_stage: float = 2
    min_width: int = 6
    max_width: int = 60
    min_height: int = 4
    max_height: int = 100
    legend_width: int = 5
    font_sizes: Dict[str, int] = None

    def __post_init__(self):
        if self.font_sizes is None:
            self.font_sizes = {
                'title': 14,
                'axis': 12,
                'tick': 10,
                'legend': 10,
                'text': 9
            }


class ModelParallelismVisualizer:

    def __init__(self, args, groups_dict, rank_models):
        """
        初始化可视化器

        Args:
            args: 并行配置参数
            groups_dict: 并行分组信息
            rank_models: 每个rank的模型配置字典
        """
        self.groups_dict = groups_dict
        self.rank_models = rank_models
        self.config = VisualizationConfig()

        # 计算布局参数
        self.layout = self._calculate_layout(args)

    def _calculate_layout(self, args) -> Dict[str, Any]:
        """计算布局参数"""
        tp_groups = self.groups_dict["tp"]
        dp_groups = self.groups_dict.get(
            "dp", [[rank] for rank in range(args.world_size)])

        dp_size = args.data_parallel_size
        dp_groups = list(zip(*dp_groups))  # 转置DP分组
        vpp_size = args.virtual_pipeline_parallel_size

        # 计算模型代表rank
        model_representatives = {}
        for dp_idx, dp_group in enumerate(dp_groups):
            for tp_idx, tp_group in enumerate(tp_groups):
                representative_rank = next(
                    (r for r in dp_group if r in tp_group), None)
                if representative_rank is not None:
                    model_representatives[(dp_idx, tp_idx)
                                          ] = representative_rank

        return {
            'num_layers': args.num_layers,
            'tp_groups': tp_groups,
            'dp_groups': dp_groups,
            'dp_size': dp_size,
            'tp_size': args.tensor_parallel_size,
            'pp_size': args.pipeline_parallel_size,
            'vpp_size': vpp_size,
            'num_tp': len(tp_groups),
            'num_model_copies': len(dp_groups),
            'model_representatives': model_representatives
        }

    def _calculate_figure_size(self) -> Tuple[float, float]:
        """动态计算图表尺寸"""
        fig_height = min(
            max(self.config.min_height,
                self.config.height_per_tp * self.layout['num_tp']),
            self.config.max_height
        )
        fig_width = min(
            max(self.config.min_width,
                self.config.width_per_stage * self.layout['vpp_size']),
            self.config.max_width
        )
        # 为图例预留0空间
        return fig_width + self.config.legend_width, fig_height

    def _setup_colors(self):
        """设置颜色配置"""
        num_copies = self.layout['num_model_copies']

        if num_copies <= 20:
            self.model_colors = plt.cm.tab20(range(num_copies))
        elif num_copies <= 40:
            # 40个以内使用tab20色板组合
            self.model_colors = np.vstack(
                [plt.cm.tab20b(range(20)), plt.cm.tab20c(range(20))])
        else:
            # 超过20个使用连续色板
            colormap = plt.cm.get_cmap('viridis')
            norm = plt.Normalize(0, num_copies - 1)
            self.model_colors = [colormap(norm(i)) for i in range(num_copies)]

    def _setup_axes(self, ax):
        """设置坐标轴"""
        ax.set_xlabel('Virtual Pipeline Stage',
                      fontsize=self.config.font_sizes['axis'])
        ax.set_ylabel('TP Group', fontsize=self.config.font_sizes['axis'])

        # 设置标题
        title = (f'Model Parallelism Configuration | '
                 f'Total Layers: {self.layout["num_layers"]} | '
                 f'DP={self.layout["dp_size"]} | '
                 f'TP={self.layout["tp_size"]} | '
                 f'PP={self.layout["pp_size"]} | '
                 f'VPP={self.layout["vpp_size"]}')
        ax.set_title(title, fontsize=self.config.font_sizes['title'], pad=20)

        # 设置坐标轴范围
        ax.set_xlim(-0.5, self.layout['vpp_size'] - 0.5)
        ax.set_ylim(-0.5, self.layout['num_tp'] - 0.5)
        ax.grid(True, linestyle='--', alpha=0.3)

        # 设置刻度
        self._setup_ticks(ax)

    def _setup_ticks(self, ax):
        """设置坐标轴刻度"""
        # X轴刻度
        ax.set_xticks(range(self.layout['vpp_size']))
        ax.set_xticklabels(
            [f'Stage {i}' for i in range(self.layout['vpp_size'])])

        # Y轴刻度（TP组标签）
        tp_labels = []
        for tp_idx, tp_group in enumerate(self.layout['tp_groups']):
            min_rank = min(tp_group)
            max_rank = max(tp_group)
            if min_rank == max_rank:
                # 单卡情况
                tp_labels.append(f'Rank{min_rank}')
            else:
                tp_labels.append(f'Group{tp_idx}: Ranks{min_rank}-{max_rank}')

        ax.set_yticks(range(self.layout['num_tp']))
        ax.set_yticklabels(tp_labels, fontsize=self.config.font_sizes['tick'])

    def _create_layer_text(self, layers: List) -> str:
        """创建层描述文本"""
        text_parts = []
        trans_layers = [l for l in layers if isinstance(l, int)]
        if "Embedding" in layers:
            text_parts.append("Embed")
        if trans_layers:
            text_parts.append(f"L{min(trans_layers)}-{max(trans_layers)}")
        if "OutputLayer" in layers:
            text_parts.append("Out")
        return "+".join(text_parts)

    def _draw_model_blocks(self, ax):
        """绘制模型块"""
        # 计算总块数
        total_blocks = sum(len(self.rank_models.get(representative_rank, []))
                           for representative_rank in self.layout['model_representatives'].values())
        if total_blocks > 5000:
            logger.warning(
                f"Large-scale drawing: {total_blocks} model stages, may take a long time")
        with tqdm(total=total_blocks, desc="Drawing model stages", unit="stage") as pbar:
            for (dp_idx, tp_idx), representative_rank in self.layout['model_representatives'].items():
                model_configs = self.rank_models.get(representative_rank, [])

                for vpp_stage, layers in enumerate(model_configs):
                    self._draw_single_block(
                        ax, dp_idx, tp_idx, vpp_stage, layers)
                    pbar.update(1)
        if total_blocks > 20000:
            logger.warning(f"Large-scale visualization: {total_blocks} model stages detected, "
                           f"this may cause memory overflow")

    def _draw_single_block(self, ax, dp_idx: int, tp_idx: int, vpp_stage: int, layers: List):
        """绘制单个模型块"""
        # 解析层信息
        layer_text = self._create_layer_text(layers)

        # 绘制矩形区块
        rect = Rectangle(
            (vpp_stage - 0.45, tp_idx - 0.45), 0.9, 0.9,
            facecolor=self.model_colors[dp_idx],
            edgecolor='white',
            linewidth=1,
            alpha=0.8
        )
        ax.add_patch(rect)

        # 添加文本标签
        fontsize = self.config.font_sizes['text']
        ax.text(
            vpp_stage, tp_idx,
            layer_text,
            ha='center', va='center',
            fontsize=fontsize,
            bbox=dict(facecolor='white', alpha=0.3, pad=1)
        )

    def _add_legend(self, ax):
        """添加图例"""
        legend_elements = [
            plt.Rectangle((0, 0), 1, 1,
                          facecolor=self.model_colors[i],
                          alpha=0.8,
                          label=f'Model Copy {i+1}')
            for i in range(self.layout['num_model_copies'])
        ]

        ax.legend(
            handles=legend_elements,
            loc='center left',  # 以左侧中心为基准
            bbox_to_anchor=(1.02, 0.5),
            title=f"Model Copies",
            title_fontsize=self.config.font_sizes['legend'],
            fontsize=self.config.font_sizes['legend']
        )

    def _visualize(self) -> plt.Figure:
        """
        执行可视化
        """
        # 创建图表
        fig_width, fig_height = self._calculate_figure_size()
        fig, ax = plt.subplots(figsize=(fig_width, fig_height), clear=True)
        self._setup_colors()
        self._setup_axes(ax)

        # 调整布局
        right_margin = 1 - self.config.legend_width / fig_width
        plt.tight_layout(rect=[0, 0, right_margin, 1])

        # 添加图例
        self._add_legend(ax)

        # 绘制模型块
        self._draw_model_blocks(ax)
        return fig

    def save(self, filepath: str):
        """
        保存可视化结果到文件
        """
        fig = self._visualize()
        logger.info(f"Saving visualization result to {filepath}...")
        fig.savefig(filepath, dpi=300, bbox_inches='tight')
        plt.close(fig)


@dataclass
class ParallelConfig:
    world_size: int
    num_layers: int
    tensor_parallel_size: int = 1
    pipeline_parallel_size: int = 1
    # 优先级高于virtual_pipeline_parallel_size
    num_layers_per_virtual_pipeline_stage: int = None
    order: str = "tp-cp-ep-dp-pp"
    standalone_embedding_stage: bool = False
    output_path: str = './'
    # _calculate_derived_parameters
    virtual_pipeline_parallel_size: int = 1
    data_parallel_size: int = None
    transformer_pipeline_parallel_size: int = None

    def __post_init__(self):
        self._validate()
        self._calculate_derived_parameters()
        self._create_output_path()

    def _calculate_derived_parameters(self):
        """计算派生参数"""
        # Pipeline模型并行大小
        self.pipeline_parallel_size = min(
            self.pipeline_parallel_size,
            (self.world_size // self.tensor_parallel_size)
        )

        self.transformer_pipeline_parallel_size = (
            self.pipeline_parallel_size - 1
            if self.standalone_embedding_stage else
            self.pipeline_parallel_size
        )

        self._validate_pipeline_configs()
        # 通过num_layers_per_virtual_pipeline_stage配置vpp
        if self.num_layers_per_virtual_pipeline_stage is not None:
            num_layers_per_pipeline_stage = self.num_layers // self.transformer_pipeline_parallel_size

            self.virtual_pipeline_parallel_size = (
                num_layers_per_pipeline_stage // self.num_layers_per_virtual_pipeline_stage
            )
            if self.virtual_pipeline_parallel_size > 1 and self.pipeline_parallel_size <= 1:
                raise ValueError(
                    "pipeline-model-parallel size should be greater than 1 with virtual-pipeline"
                )

        # 计算数据并行大小
        self.data_parallel_size = self.world_size // (
            self.tensor_parallel_size * self.pipeline_parallel_size
        )

    def _validate_pipeline_configs(self):
        """验证流水线层数配置"""
        if self.transformer_pipeline_parallel_size == 0 or \
                self.num_layers % self.transformer_pipeline_parallel_size != 0:
            if self.standalone_embedding_stage:
                raise ValueError(
                    f'number of layers ({self.num_layers}) must be divisible by '
                    f'transformer_pipeline_parallel_size ({self.transformer_pipeline_parallel_size}), '
                    f'which is pipeline_parallel_size-1 when standalone_embedding_stage is enabled.'
                )
            else:
                raise ValueError(
                    f'number of layers ({self.num_layers}) must be divisible by '
                    f'pipeline_parallel_size ({self.pipeline_parallel_size}).'
                )

        num_layers_per_pipeline_stage = self.num_layers // self.transformer_pipeline_parallel_size
        if self.num_layers_per_virtual_pipeline_stage is not None:
            if num_layers_per_pipeline_stage % self.num_layers_per_virtual_pipeline_stage != 0:
                raise ValueError(
                    f'number of layers per pipeline stage ({num_layers_per_pipeline_stage}) must be divisible by '
                    f'number of layers per virtual pipeline stage ({self.num_layers_per_virtual_pipeline_stage}).'
                )

    def _create_output_path(self):
        create_directory(self.output_path)
        file_name = f"ws{self.world_size}_ln{self.num_layers}" \
                    f"_tp{self.tensor_parallel_size}" \
                    f"_pp{self.pipeline_parallel_size}" \
                    f"_vpp{self.virtual_pipeline_parallel_size}.png"
        png_path = os.path.join(self.output_path, file_name)

        if os.path.exists(png_path):
            logger.warning(f"Existing path will be recovered: {png_path}")
            remove_path(png_path)
        self.output_path = png_path

    def _validate(self):
        """
        验证所有配置输入的类型
        order 会在RankGroupGenerator类中验证
        data_parallel_size 会在 _calculate_derived_parameters中计算覆盖
        """
        per_vpp = self.num_layers_per_virtual_pipeline_stage if \
            self.num_layers_per_virtual_pipeline_stage is not None else 1
        if self.tensor_parallel_size is None:
            self.tensor_parallel_size = 1
        if self.pipeline_parallel_size is None:
            self.pipeline_parallel_size = 1
        for name, _input in {
            "world_size": self.world_size,
            "num_layers": self.num_layers,
            "tensor_parallel_size": self.tensor_parallel_size,
            "pipeline_parallel_size": self.pipeline_parallel_size,
            "num_layers_per_virtual_pipeline_stage": per_vpp
        }.items():
            if not is_int(_input) or _input < 1:
                raise ValueError(f"{name} must be a positive integer")
        if self.world_size > 1024:
            raise ValueError(
                "world_size ({self.world_size}) exceeds the maximum supported value of 1024.")
        if self.num_layers > 256:
            raise ValueError(
                "num_layers ({self.num_layers}) exceeds the maximum supported value of 256.")

        if (self.world_size % (self.tensor_parallel_size * self.pipeline_parallel_size) != 0):
            raise ValueError(
                f"world_size ({self.world_size}) is not divisible by "
                f"tensor_parallel_size({self.tensor_parallel_size}) x "
                f"pipeline_parallel_size ({self.pipeline_parallel_size})"
            )
        if not isinstance(self.standalone_embedding_stage, bool):
            raise ValueError(f"standalone_embedding_stage must be a boolean")


class ModelLayerSimulator:
    """
    模型仿真 - 统一处理模型层信息生成
    """

    def __init__(self, args: 'ParallelConfig', pp_stage_list: list):
        self.args = args
        self.pp_stage_list = pp_stage_list

    def get_model_for_rank(self, rank: int) -> List[List[str]]:
        """
        为指定rank生成模型层
        """
        if self._should_use_virtual_pipeline():
            return self._create_virtual_pipeline_model(rank)
        else:
            return self._create_standard_pipeline_model(rank)

    def _should_use_virtual_pipeline(self) -> bool:
        """判断是否使用虚拟流水线策略"""
        return self.args.virtual_pipeline_parallel_size > 1

    def _create_virtual_pipeline_model(self, rank: int) -> List[List[str]]:
        """虚拟流水线模式：生成多个虚拟阶段的模型配置"""
        model_configs = []
        for vpp_stage in range(self.args.virtual_pipeline_parallel_size):
            # 确定当前阶段是否为流水线的首尾阶段
            is_first_stage = self._is_pipeline_first_stage(rank, vpp_stage)
            is_last_stage = self._is_pipeline_last_stage(rank, vpp_stage)

            # 生成单个虚拟阶段的模型
            stage_config = self._create_single_stage_config(
                rank, vpp_stage, is_first_stage, is_last_stage
            )
            model_configs.append(stage_config)

        return model_configs

    def _create_standard_pipeline_model(self, rank: int) -> List[List[str]]:
        """标准流水线模式：生成单个模型配置"""
        is_first_stage = self._is_pipeline_first_stage(rank, 0)
        is_last_stage = self._is_pipeline_last_stage(rank, -1)

        model_config = self._create_single_stage_config(
            rank, 0, is_first_stage, is_last_stage
        )
        return [model_config]

    def _create_single_stage_config(self, rank: int, vpp_stage: int,
                                    is_first_stage: bool, is_last_stage: bool) -> List[str]:
        """
        生成单个阶段（虚拟阶段或普通阶段）的模型配置
        """
        components = []

        # 前处理组件（嵌入层）
        if is_first_stage:
            components.append("Embedding")
        # Transformer层组件
        components.extend(self._create_transformer_layers(rank, vpp_stage))
        # 后处理组件（输出层）
        if is_last_stage:
            components.append("OutputLayer")

        return components

    def _create_transformer_layers(self, rank: int, vpp_stage: int) -> List[str]:
        """创建Transformer层组件"""
        num_layers = self._calculate_transformer_layers(rank)
        if num_layers == 0:
            return []

        # 计算层偏移量
        offset = self._calculate_layer_offset(rank, vpp_stage)

        # 生成层名称
        return [i + 1 + offset for i in range(num_layers)]

    def _calculate_transformer_layers(self, rank: int) -> int:
        """计算当前rank需要处理的Transformer层数"""
        # 流水线并行场景
        if self.args.pipeline_parallel_size > 1:
            if (self.args.standalone_embedding_stage and
                    self._is_pipeline_first_stage(rank, 0)):
                return 0  # 嵌入阶段不包含Transformer层
            num_layers = self.args.num_layers // self.args.transformer_pipeline_parallel_size
            if self._should_use_virtual_pipeline():
                return num_layers // self.args.virtual_pipeline_parallel_size
            return num_layers
        # 非流水线并行场景
        return self.args.num_layers

    def _calculate_layer_offset(self, rank: int, vpp_stage: int) -> int:
        """计算层偏移量"""
        pipeline_rank = self._get_pipeline_rank(rank)
        if self.args.standalone_embedding_stage:
            pipeline_rank -= 1
        base_layers_per_stage = self.args.num_layers // self.args.transformer_pipeline_parallel_size

        if self._should_use_virtual_pipeline():
            # 虚拟流水线：进一步分割每个流水线阶段
            layers_per_virtual_stage = base_layers_per_stage // self.args.virtual_pipeline_parallel_size
            return vpp_stage * (self.args.num_layers // self.args.virtual_pipeline_parallel_size) + \
                (pipeline_rank * layers_per_virtual_stage)
        else:
            # 普通流水线：直接按流水线阶段划分
            return pipeline_rank * base_layers_per_stage

    def _is_pipeline_first_stage(self, rank: int, vpp_stage: int) -> bool:
        """判断是否为流水线首阶段"""
        return (rank in self.pp_stage_list[0] and vpp_stage == 0)

    def _is_pipeline_last_stage(self, rank: int, vpp_stage: int) -> bool:
        """判断是否为流水线尾阶段"""
        # 虚拟流水线：只有最后一个虚拟阶段才算是真正的尾阶段
        if (self._should_use_virtual_pipeline() and
                vpp_stage != (self.args.virtual_pipeline_parallel_size - 1)):
            return False

        return rank in self.pp_stage_list[-1]

    def _get_pipeline_rank(self, rank: int) -> int:
        """获取在流水线中的局部rank"""
        for local_rank, ranks in enumerate(self.pp_stage_list):
            if rank in ranks:
                return local_rank
        return 0

    # 批量生成所有rank的模型配置
    def generate_all_models(self) -> Dict[int, List[List[str]]]:
        """批量生成所有rank的模型配置"""
        return {rank: self.get_model_for_rank(rank) for rank in range(self.args.world_size)}


def plot_model_parallelism(args: 'ParallelConfig'):
    logger.info("Starting model parallelism visualization")
    logger.info("Generating model configurations for each rank...")
    groups_dict = RankGroupGenerator(
        args.tensor_parallel_size,
        1,  # expert_parallel
        args.data_parallel_size,
        args.pipeline_parallel_size,
        1,  # context_parallel
        args.order
    ).generate_all_ranks()
    pp_stage_list = list(zip(*groups_dict.get("pp", [])))
    rank_models = ModelLayerSimulator(
        args, pp_stage_list).generate_all_models()
    logger.info(f"Model configuration generation completed: "
                f"{len(groups_dict.get('tp', []))} TP groups, "
                f"{len(groups_dict.get('dp', []))} DP groups, "
                f"{len(groups_dict.get('pp', []))} PP groups")
    logger.info("Starting visualization...")
    ModelParallelismVisualizer(
        args, groups_dict, rank_models).save(args.output_path)
    logger.info("Done")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import math
import os
import sys
from collections import namedtuple

import torch
import pandas as pd

from msprobe.core.common.file_utils import write_csv, read_csv
from msprobe.pytorch.api_accuracy_checker.common.config import msCheckerConfig
from msprobe.pytorch.api_accuracy_checker.compare.compare_utils import API_PRECISION_COMPARE_RESULT_FILE_NAME, \
    API_PRECISION_COMPARE_DETAILS_FILE_NAME, BENCHMARK_COMPARE_SUPPORT_LIST, API_PRECISION_COMPARE_UNSUPPORT_LIST, \
    ApiPrecisionCompareColumn, absolute_standard_api, binary_standard_api, ulp_standard_api, thousandth_standard_api, \
    BINARY_COMPARE_UNSUPPORT_LIST, ULP_COMPARE_SUPPORT_LIST, convert_str_to_float, CompareMessage
from msprobe.pytorch.api_accuracy_checker.compare.compare_input import PrecisionCompareInput
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_register import StandardRegistry
from msprobe.pytorch.api_accuracy_checker.precision_standard.ulp_compare import UlpPrecisionCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.benchmark_compare import BenchmarkPrecisionCompare
from msprobe.pytorch.api_accuracy_checker.precision_standard.standard_config import StandardConfig
from msprobe.pytorch.api_accuracy_checker.compare.compare_column import ApiPrecisionOutputColumn
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import get_validated_result_csv_path
from msprobe.pytorch.api_accuracy_checker.common.utils import extract_detailed_api_segments, extract_basic_api_segments
from msprobe.core.common.file_utils import FileChecker, change_mode, create_directory
from msprobe.pytorch.common.log import logger
from msprobe.core.common.utils import CompareException, check_op_str_pattern_valid
from msprobe.core.common.const import Const, CompareConst, FileCheckConst

CompareConfig = namedtuple('CompareConfig', ['npu_csv_path', 'gpu_csv_path', 'result_csv_path', 'details_csv_path'])
BenchmarkInfNanConsistency = namedtuple('BenchmarkInfNanConsistency', ['small_value_inf_nan_consistency', 
                                                                           'rmse_inf_nan_consistency', 
                                                                           'max_rel_inf_nan_consistency', 
                                                                           'mean_rel_inf_nan_consistency', 
                                                                           'eb_inf_nan_consistency'])
UNSUPPORTED_MESSAGE = 'This data type does not support benchmark compare.'


benchmark_message = {
    "small_value_err_status": {
        CompareConst.ERROR: "ERROR: 小值域错误比值超过阈值\n",
        CompareConst.WARNING: "WARNING: 小值域错误比值超过阈值\n"
    },
    "rmse_status": {
        CompareConst.ERROR: "ERROR: 均方根误差比值超过阈值\n",
        CompareConst.WARNING: "WARNING: 均方根误差比值超过阈值\n"
    },
    "max_rel_err_status": {
        CompareConst.ERROR: "ERROR: 相对误差最大值比值超过阈值\n",
        CompareConst.WARNING: "WARNING: 相对误差最大值比值超过阈值\n"
    },
    "mean_rel_err_status": {
        CompareConst.ERROR: "ERROR: 相对误差平均值比值超过阈值\n",
        CompareConst.WARNING: "WARNING: 相对误差平均值比值超过阈值\n"
    }
}


def write_detail_csv(content, save_path):
    rows = []
    content = ["{:.{}f}".format(item, msCheckerConfig.precision) \
                   if isinstance(item, float) else item for item in content]
    rows.append(content)
    write_csv(rows, save_path)


def register_compare_func():
    registry = StandardRegistry()
    registry.register(CompareConst.ABSOLUTE_THRESHOLD, record_absolute_threshold_result)
    registry.register(CompareConst.BINARY_CONSISTENCY, record_binary_consistency_result)
    registry.register(CompareConst.ULP_COMPARE, record_ulp_compare_result)
    registry.register(CompareConst.THOUSANDTH_STANDARD, record_thousandth_threshold_result)
    registry.register(CompareConst.BENCHMARK, record_benchmark_compare_result)
    registry.register(CompareConst.ACCUMULATIVE_ERROR_COMPARE, record_accumulative_error_compare_result)
    return registry


def api_precision_compare(config):
    logger.info("Start compare task")
    logger.info(f"Compare task result will be saved in {config.result_csv_path}")
    logger.info(f"Compare task detail will be saved in {config.details_csv_path}")
    try:
        npu_data = read_csv(config.npu_csv_path)
    except Exception as err:
        logger.error(f"Open npu csv Error: %s" % str(err))
    check_csv_columns(npu_data.columns, "npu_csv")
    try:
        gpu_data = read_csv(config.gpu_csv_path)
    except Exception as err:
        logger.error(f"Open gpu csv Error: %s" % str(err))
    check_csv_columns(gpu_data.columns, "gpu_csv")
    detail_csv_title = [ApiPrecisionCompareColumn.get_detail_csv_title()]
    result_csv_title = [ApiPrecisionCompareColumn.get_result_csv_title()]
    write_csv(result_csv_title, config.result_csv_path)
    write_csv(detail_csv_title, config.details_csv_path)
    try:
        analyse_csv(npu_data, gpu_data, config)
    except Exception as err:
        logger.error(f"Analyse csv Error: %s" % str(err))
    change_mode(config.result_csv_path, FileCheckConst.DATA_FILE_AUTHORITY)
    change_mode(config.details_csv_path, FileCheckConst.DATA_FILE_AUTHORITY)


def analyse_csv(npu_data, gpu_data, config):
    forward_status, backward_status = [], []
    last_api_name, last_api_dtype, last_api_full_name = None, None, None
    last_api_skip_message = ''
    registry = register_compare_func()
    
    for _, row_npu in npu_data.iterrows():
        message = ''
        compare_column = ApiPrecisionOutputColumn()
        full_api_name_with_direction_status = row_npu[ApiPrecisionCompareColumn.API_NAME]
        check_op_str_pattern_valid(full_api_name_with_direction_status)
        row_gpu = gpu_data[gpu_data[ApiPrecisionCompareColumn.API_NAME] == full_api_name_with_direction_status]
        api_name, api_full_name, direction_status = extract_detailed_api_segments(full_api_name_with_direction_status)
        if not api_full_name:
            err_message = f"The API name {full_api_name_with_direction_status} is invalid."
            logger.error(err_message)
            compare_column.api_name = full_api_name_with_direction_status
            compare_column.compare_result = CompareConst.SKIP
            compare_column.compare_message = err_message
            write_detail_csv(compare_column.to_column_value(), config.details_csv_path)
            write_csv([[full_api_name_with_direction_status, CompareConst.SKIP, CompareConst.SKIP, err_message]], 
                      config.result_csv_path)
            continue
        if row_gpu.empty:
            logger.warning(f'This API : {full_api_name_with_direction_status} does not exist in the GPU data.')
            continue
        if len(row_gpu) > 1:
            msg = f'This API : {full_api_name_with_direction_status} has multiple records in the GPU data.'
            raise CompareException(CompareException.INVALID_DATA_ERROR, msg)
        row_gpu = row_gpu.iloc[0]
        new_status = CompareConst.SPACE
        try:
            new_status = get_api_status(row_npu, row_gpu, api_name, compare_column, registry)
        except Exception as err:
            logger.error(f"Get api status error: {str(err)}")
            compare_column.api_name = full_api_name_with_direction_status
            compare_column.compare_result = CompareConst.SKIP
            compare_column.compare_message = str(err)
            write_detail_csv(compare_column.to_column_value(), config.details_csv_path)
            write_csv([[full_api_name_with_direction_status, CompareConst.SKIP, CompareConst.SKIP, str(err)]], 
                      config.result_csv_path)
            continue
        
        write_detail_csv(compare_column.to_column_value(), config.details_csv_path)

        if last_api_name is not None and api_full_name != last_api_name:
            if last_api_dtype in API_PRECISION_COMPARE_UNSUPPORT_LIST:
                message = UNSUPPORTED_MESSAGE
                write_csv([[last_api_name, CompareConst.SKIP, CompareConst.SKIP, message]], config.result_csv_path)
                print_test_success(last_api_name, CompareConst.SKIP, CompareConst.SKIP)
            else:
                forward_result = get_api_checker_result(forward_status)
                backward_result = get_api_checker_result(backward_status)
                _, base_api_name = extract_basic_api_segments(last_api_name)
                message += CompareMessage.get(base_api_name, "") if forward_result == CompareConst.ERROR else ""
                message += last_api_skip_message if forward_result == CompareConst.SKIP else ""
                write_csv([[last_api_name, forward_result, backward_result, message]], config.result_csv_path)
                print_test_success(last_api_name, forward_result, backward_result)
                last_api_skip_message = ''
            forward_status, backward_status = [], []
            message = ''

        is_supported = row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE] not in API_PRECISION_COMPARE_UNSUPPORT_LIST
        last_api_name = api_full_name

        last_api_dtype = row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE]
        if not is_supported:
            continue

        if direction_status == 'forward':
            forward_status.append(new_status)
            last_api_skip_message = str(row_npu[ApiPrecisionCompareColumn.MESSAGE]) if new_status == CompareConst.SKIP \
                                    else ''
        elif direction_status == 'backward':
            backward_status.append(new_status)
        else:
            logger.error(f"Invalid direction status: {direction_status}")

    if last_api_name is not None:
        if last_api_dtype in API_PRECISION_COMPARE_UNSUPPORT_LIST:
            message = UNSUPPORTED_MESSAGE
            write_csv([[last_api_name, CompareConst.SKIP, CompareConst.SKIP, message]], config.result_csv_path)
            print_test_success(last_api_name, CompareConst.SKIP, CompareConst.SKIP)
        else:
            forward_result = get_api_checker_result(forward_status)
            backward_result = get_api_checker_result(backward_status)
            _, base_api_name = extract_basic_api_segments(last_api_name)
            message += CompareMessage.get(base_api_name, "") if forward_result == CompareConst.ERROR else ""
            message += last_api_skip_message if forward_result == CompareConst.SKIP else ""
            write_csv([[last_api_name, forward_result, backward_result, message]], config.result_csv_path)
            print_test_success(last_api_name, forward_result, backward_result)
            last_api_skip_message = ''


def get_api_status(row_npu, row_gpu, api_name, compare_column, registry):
    full_api_name_with_direction_status = row_npu[ApiPrecisionCompareColumn.API_NAME]
    # 当前API的输出为空（例如反向过程中requires_grad=False）,跳过比对
    if row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE].isspace() or \
        row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE] in API_PRECISION_COMPARE_UNSUPPORT_LIST or \
        row_npu[ApiPrecisionCompareColumn.SHAPE] == CompareConst.ZERO_SHAPE:
        compare_column.api_name = full_api_name_with_direction_status
        compare_column.compare_result = CompareConst.SKIP
        compare_column.compare_message = row_npu[ApiPrecisionCompareColumn.MESSAGE]
        new_status = CompareConst.SKIP
    else:
        compare_column.api_name = full_api_name_with_direction_status
        dtype = row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE]
        input_data = PrecisionCompareInput(row_npu, row_gpu, dtype, compare_column, api_name=api_name)
        in_dtype = get_in_dtype(row_npu)
        comparison_func = registry.get_comparison_function(api_name, dtype, in_dtype)
        new_status = comparison_func(input_data)
    return new_status


def get_in_dtype(row_npu):
    if row_npu[ApiPrecisionCompareColumn.REL_ERR_RATIO].isspace():
        return "torch.float32"
    else:
        return "torch.float8_e4m3fn"


def print_test_success(api_full_name, forward_result, backward_result):
    is_fwd_success = (forward_result == CompareConst.PASS)
    is_bwd_success = (backward_result == CompareConst.PASS or backward_result == CompareConst.SPACE)
    logger.info(f"running api_full_name {api_full_name} compare, "
                f"is_fwd_success: {is_fwd_success}, "
                f"is_bwd_success: {is_bwd_success}")


def check_error_rate(npu_error_rate):
    return CompareConst.PASS if convert_str_to_float(npu_error_rate) == 0 else CompareConst.ERROR


def get_absolute_threshold_result(row_npu):
    inf_nan_error_ratio = convert_str_to_float(row_npu[ApiPrecisionCompareColumn.INF_NAN_ERROR_RATIO])
    rel_err_ratio = convert_str_to_float(row_npu[ApiPrecisionCompareColumn.REL_ERR_RATIO])
    abs_err_ratio = convert_str_to_float(row_npu[ApiPrecisionCompareColumn.ABS_ERR_RATIO])

    inf_nan_result = CompareConst.PASS if inf_nan_error_ratio == 0 else CompareConst.ERROR
    rel_err_result = CompareConst.PASS if rel_err_ratio == 0 else CompareConst.ERROR
    abs_err_result = CompareConst.PASS if abs_err_ratio == 0 else CompareConst.ERROR

    if CompareConst.ERROR in [inf_nan_result, rel_err_result, abs_err_result]:
        absolute_threshold_result = CompareConst.ERROR
    else:
        absolute_threshold_result = CompareConst.PASS

    return {
        "inf_nan_error_ratio": inf_nan_error_ratio,
        "inf_nan_result": inf_nan_result,
        "rel_err_ratio": rel_err_ratio,
        "rel_err_result": rel_err_result,
        "abs_err_ratio": abs_err_ratio,
        "abs_err_result": abs_err_result,
        "absolute_threshold_result": absolute_threshold_result,
    }


def get_api_checker_result(status):
    if not status:
        return CompareConst.SPACE
    if all(item == CompareConst.SKIP for item in status):
        return CompareConst.SKIP
    for const in (CompareConst.ERROR, CompareConst.WARNING):
        if const in status:
            return const
    return CompareConst.PASS


def check_csv_columns(columns, csv_type):
    required_columns = ApiPrecisionCompareColumn.to_required_columns()
    missing_columns = [column for column in required_columns if column not in columns]
    if missing_columns:
        msg = f"The following columns {','.join(missing_columns)} are missing in{csv_type}"
        raise CompareException(CompareException.INVALID_DATA_ERROR, msg)


def record_binary_consistency_result(input_data):
    row_npu = input_data.row_npu
    compare_column = input_data.compare_column
    new_status = check_error_rate(row_npu[ApiPrecisionCompareColumn.ERROR_RATE])
    compare_column.error_rate = row_npu[ApiPrecisionCompareColumn.ERROR_RATE]
    compare_column.error_rate_status = new_status
    compare_column.compare_result = new_status
    compare_column.compare_algorithm = CompareConst.BINARY_CONSISTENCY_ALGORITHM_NAME
    message = ''
    if compare_column.error_rate_status == CompareConst.ERROR:
        message += "ERROR: 二进制一致错误率超过阈值\n"
    compare_column.compare_message = message
    return new_status


def record_absolute_threshold_result(input_data):
    row_npu = input_data.row_npu
    compare_column = input_data.compare_column
    absolute_threshold_result = get_absolute_threshold_result(row_npu)
    compare_column.inf_nan_error_ratio = absolute_threshold_result.get("inf_nan_error_ratio")
    compare_column.inf_nan_error_ratio_status = absolute_threshold_result.get("inf_nan_result")
    compare_column.rel_err_ratio = absolute_threshold_result.get("rel_err_ratio")
    compare_column.rel_err_ratio_status = absolute_threshold_result.get("rel_err_result")
    compare_column.abs_err_ratio = absolute_threshold_result.get("abs_err_ratio")
    compare_column.abs_err_ratio_status = absolute_threshold_result.get("abs_err_result")
    compare_column.compare_result = absolute_threshold_result.get("absolute_threshold_result")
    compare_column.compare_algorithm = CompareConst.ABSOLUTE_THRESHOLD_ALGORITHM_NAME
    message = ''
    if compare_column.inf_nan_error_ratio_status == CompareConst.ERROR:
        message += "ERROR: inf/nan错误率超过阈值"
    if compare_column.rel_err_ratio_status == CompareConst.ERROR:
        message += "ERROR: 相对误差错误率超过阈值"
    if compare_column.abs_err_ratio_status == CompareConst.ERROR:
        message += "ERROR: 绝对误差错误率超过阈值"
    compare_column.compare_message = message
    return compare_column.compare_result


def record_benchmark_compare_result(input_data):
    bs = BenchmarkPrecisionCompare(input_data)
    compare_result = bs.compare()
    for status_attr, messages in benchmark_message.items():
        status_value = getattr(input_data.compare_column, status_attr)
        if status_value in messages:
            input_data.compare_column.compare_message += messages[status_value]
    return compare_result


def record_ulp_compare_result(input_data):
    us = UlpPrecisionCompare(input_data)
    compare_result = us.compare()
    return compare_result


def record_accumulative_error_compare_result(input_data):
    row_npu = input_data.row_npu
    compare_column = input_data.compare_column
    absolute_threshold_result = get_absolute_threshold_result(row_npu)
    threshold_result = absolute_threshold_result.get("absolute_threshold_result")
    eb, eb_result = check_eb(row_npu)
    accumulative_error_compare_result = CompareConst.PASS
    if CompareConst.ERROR in [threshold_result, eb_result]:
        accumulative_error_compare_result = CompareConst.ERROR
    
    compare_column.inf_nan_error_ratio = absolute_threshold_result.get("inf_nan_error_ratio")
    compare_column.inf_nan_error_ratio_status = absolute_threshold_result.get("inf_nan_result")
    compare_column.rel_err_ratio = absolute_threshold_result.get("rel_err_ratio")
    compare_column.rel_err_ratio_status = absolute_threshold_result.get("rel_err_result")
    compare_column.abs_err_ratio = absolute_threshold_result.get("abs_err_ratio")
    compare_column.abs_err_ratio_status = absolute_threshold_result.get("abs_err_result")
    compare_column.eb_ratio = eb
    compare_column.eb_status = eb_result
    compare_column.compare_result = accumulative_error_compare_result
    compare_column.compare_algorithm = CompareConst.ACCUMULATIVE_ERROR_COMPARE_ALGORITHM_NAME
    message = []
    if compare_column.inf_nan_error_ratio_status == CompareConst.ERROR:
        message.append("ERROR: inf/nan错误率超过阈值\n")
    if compare_column.rel_err_ratio_status == CompareConst.ERROR:
        message.append("ERROR: 相对误差错误率超过阈值\n")
    if compare_column.abs_err_ratio_status == CompareConst.ERROR:
        message.append("ERROR: 绝对误差错误率超过阈值\n")
    if compare_column.eb_status == CompareConst.ERROR:
        message.append("ERROR: 误差均衡性超过阈值\n")
    compare_column.compare_message = "\n".join(message)
    return compare_column.compare_result


def check_eb(row_npu):
    eb = convert_str_to_float(row_npu[ApiPrecisionCompareColumn.EB])
    dtype = row_npu[ApiPrecisionCompareColumn.DEVICE_DTYPE]
    eb_threshold = StandardConfig.get_accumulative_error_eb_threshold(dtype)
    eb_result = CompareConst.PASS if eb <= eb_threshold else CompareConst.ERROR
    return eb, eb_result


def check_thousandth_rate(thousandth_rate):
    return CompareConst.PASS if convert_str_to_float(thousandth_rate) >= CompareConst.THOUSANDTH_PASS_VALUE \
        else CompareConst.ERROR


def record_thousandth_threshold_result(input_data):
    row_npu = input_data.row_npu
    compare_column = input_data.compare_column
    new_status = check_thousandth_rate(row_npu[ApiPrecisionCompareColumn.REL_ERR_THOUSANDTH])
    compare_column.rel_err_thousandth = row_npu[ApiPrecisionCompareColumn.REL_ERR_THOUSANDTH]
    compare_column.rel_err_thousandth_status = new_status
    compare_column.compare_result = new_status
    compare_column.compare_algorithm = CompareConst.THOUSANDTH_STANDARD_ALGORITHM_NAME
    message = ''
    if compare_column.rel_err_thousandth_status == CompareConst.ERROR:
        message += "ERROR: 双千指标不达标\n"
    compare_column.compare_message = message
    return compare_column.compare_result


def _api_precision_compare(parser=None):
    if not parser:
        parser = argparse.ArgumentParser()
    _api_precision_compare_parser(parser)
    args = parser.parse_args(sys.argv[1:])
    _api_precision_compare_command(args)
    logger.info("Compare task completed.")


def _api_precision_compare_command(args):
    npu_csv_path = get_validated_result_csv_path(args.npu_csv_path, 'detail')
    gpu_csv_path = get_validated_result_csv_path(args.gpu_csv_path, 'detail')
    out_path = args.out_path if args.out_path else Const.DEFAULT_PATH
    create_directory(out_path)
    out_path_checker = FileChecker(out_path, FileCheckConst.DIR, ability=FileCheckConst.WRITE_ABLE)
    out_path = out_path_checker.common_check()
    result_csv_path = os.path.join(out_path, API_PRECISION_COMPARE_RESULT_FILE_NAME)
    details_csv_path = os.path.join(out_path, API_PRECISION_COMPARE_DETAILS_FILE_NAME)
    compare_config = CompareConfig(npu_csv_path, gpu_csv_path, result_csv_path, details_csv_path)
    api_precision_compare(compare_config)


def _api_precision_compare_parser(parser):
    parser.add_argument("-npu", "--npu_csv_path", dest="npu_csv_path", default="", type=str,
                        help="<Required> , Accuracy_checking_details.csv generated on the NPU by using the "
                             "api_accuracy_checker tool.",
                        required=True)
    parser.add_argument("-gpu", "--gpu_csv_path", dest="gpu_csv_path", default="", type=str,
                        help="<Required> Accuracy_checking_details.csv generated on the GPU by using the "
                             "api_accuracy_checker tool.",
                        required=True)
    parser.add_argument("-o", "--out_path", dest="out_path", default="", type=str,
                        help="<optional> The api precision compare task result out path.",
                        required=False)



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves the accuracy_network_compare function.
"""
import csv
import os
import stat
import re
import sys
import subprocess

import numpy as np

from msprobe.core.common.log import logger
from msprobe.infer.offline.compare.msquickcmp.common.utils import check_file_size_valid, AccuracyCompareException, \
    ACCURACY_COMPARISON_NET_OUTPUT_ERROR, ACCURACY_COMPARISON_INVALID_DATA_ERROR, MAX_READ_FILE_SIZE_4G
from msprobe.infer.utils.file_open_check import sanitize_csv_value
from msprobe.infer.utils.file_open_check import ms_open
from msprobe.infer.utils.check.rule import Rule
from msprobe.infer.utils.util import load_file_to_read_common_check, filter_cmd
from msprobe.infer.utils.constants import TENSOR_MAX_SIZE
from msprobe.msaccucmp import msaccucmp

INFO_FLAG = "[INFO]"
WRITE_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR
READ_WRITE_FLAGS = os.O_RDWR | os.O_CREAT | os.O_TRUNC
# index of each member in compare result_*.csv file
NPU_DUMP_TAG = "NPUDump"
GROUND_TRUTH_TAG = "GroundTruth"
MIN_ELEMENT_NUM = 3


class NetCompare(object):
    """
    Class for compare the entire network
    """

    def __init__(self, npu_dump_data_path, cpu_dump_data_path, output_json_path, arguments, golden_json_path=None):
        self.npu_dump_data_path = npu_dump_data_path
        self.cpu_dump_data_path = cpu_dump_data_path
        self.output_json_path = output_json_path
        self.golden_json_path = golden_json_path
        self.arguments = arguments
        self.python_version = sys.executable.split('/')[-1]

        if self.golden_json_path:
            check_file_size_valid(self.golden_json_path, MAX_READ_FILE_SIZE_4G)

    @staticmethod
    def execute_command_line(cmd):
        logger.info(f"Execute command:{' '.join(cmd)}")
        process = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        return process

    @staticmethod
    def _catch_compare_result(log_line, catch):
        result = []
        header = []
        try:
            if catch:
                # get the compare result
                info = log_line.decode().split(INFO_FLAG)
                if len(info) > 1:
                    info_content = info[1].strip().split(" ")
                    info_content = [item for item in info_content if item != '']
                    pattern_num = re.compile(r'^([0-9]+)\.?([0-9]+)?')
                    pattern_nan = re.compile(r'NaN', re.I)
                    pattern_header = re.compile(r'Cosine|Error|Distance|Divergence|Deviation', re.I)
                    match = pattern_num.match(info_content[0])
                    if match:
                        result = info_content
                    if not match and pattern_nan.match(info_content[0]):
                        result = info_content
                    if not match and pattern_header.search(info_content[0]):
                        header = info_content
            return result, header
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            logger.error('Failed to parse the alg compare result!')
            raise AccuracyCompareException(ACCURACY_COMPARISON_NET_OUTPUT_ERROR) from error
        finally:
            pass

    @staticmethod
    def _process_result_one_line(fp_write, fp_read, npu_file_name, golden_file_name, result):
        writer = csv.writer(fp_write)
        # write header to file
        table_header_info = next(fp_read)
        header_list = table_header_info.strip().split(',')
        writer.writerow(header_list)
        npu_dump_index = header_list.index(NPU_DUMP_TAG)
        ground_truth_index = header_list.index(GROUND_TRUTH_TAG)

        result_reader = csv.reader(fp_read)
        # update result data
        new_content = []
        for line in result_reader:
            if len(line) < MIN_ELEMENT_NUM:
                logger.warning(f"The content of line is {line}")
                continue
            if line[npu_dump_index] != "Node_Output":
                for ele in line:
                    sanitize_csv_value(ele)
                writer.writerow(line)
            else:
                new_content = [
                    line[0], "NaN", "Node_Output", "NaN", "NaN",
                    npu_file_name, "NaN", golden_file_name, "[]"
                ]
                new_content.extend(result)
                new_content.extend([""])
                if line[ground_truth_index] != "*":
                    writer.writerow(line)
        writer.writerow(new_content)

    @staticmethod
    def _process_result_to_csv(fp_write, csv_info):
        writer = csv.writer(fp_write)
        if csv_info.header:
            header_base_info = [
                'Index', 'OpType', 'NPUDump', 'DataType', 'Address',
                'GroundTruth', 'DataType', 'TensorIndex', 'Shape'
            ]
            header_base_info.extend(csv_info.header)
            writer.writerow(header_base_info)
        fp_write.seek(0, 0)
        index = len(fp_write.readlines()) - 1
        new_content = [
            str(index), "NaN", "Node_Output", "NaN", "NaN",
            csv_info.npu_file_name, "NaN", csv_info.golden_file_name, "[]"
        ]
        new_content.extend(csv_info.result)
        writer.writerow(new_content)

    def accuracy_network_compare(self):
        """
        Function Description:
            invoke the interface for network-wide comparison
        Exception Description:
            when invalid msaccucmp command throw exception
        """
        msaccucmp_cmd = [
            self.python_version, msaccucmp.__file__, "compare", "-m",
            self.npu_dump_data_path, "-g",
            self.cpu_dump_data_path, "-f", self.output_json_path, "-out", self.arguments.output_path
        ]

        if self.golden_json_path is not None:
            msaccucmp_cmd.extend(["-cf", self.golden_json_path])

        msaccucmp_cmd = filter_cmd(msaccucmp_cmd)
        status_code, _, _ = self.execute_msaccucmp_command(msaccucmp_cmd)
        if status_code == 2 or status_code == 0:
            logger.info(f"Finish compare the files in directory {self.npu_dump_data_path} "
                        f"with those in directory {self.cpu_dump_data_path}.")
        else:
            logger.error(f"Failed to execute command: {' '.join(msaccucmp_cmd)}")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DATA_ERROR)

    def net_output_compare(self, npu_net_output_data_path, golden_net_output_info):
        """
        net_output_compare
        """
        if not golden_net_output_info:
            return
        npu_dump_file = {}
        file_index = 0
        logger.info("=================================compare Node_output=================================")
        logger.info("start to compare the Node_output at now, compare result is:")
        logger.warning("The comparison of Node_output may be incorrect in certain scenarios. "
                       "If the precision is abnormal, "
                       "please check whether the mapping between the comparison data is correct.")
        for dir_path, _, files in os.walk(npu_net_output_data_path):
            for each_file in sorted(files):
                if each_file.endswith(".npy"):
                    npu_dump_file[file_index] = os.path.join(dir_path, each_file)
                    npu_dump_file[file_index] = load_file_to_read_common_check(npu_dump_file.get(file_index))
                    npu_data = np.load(npu_dump_file.get(file_index))
                    golden_net_output_info[file_index] = \
                            load_file_to_read_common_check(golden_net_output_info.get(file_index))
                    golden_data = np.load(golden_net_output_info.get(file_index))
                    np.save(npu_dump_file.get(file_index), npu_data.reshape(golden_data.shape))
                    msaccucmp_cmd = [
                        self.python_version, self.msaccucmp_command_file_path, "compare", "-m",
                        npu_dump_file.get(file_index), "-g", golden_net_output_info.get(file_index)
                    ]
                    status, compare_result, header = self.execute_msaccucmp_command(msaccucmp_cmd, True)
                    if status == 2 or status == 0:
                        self.save_net_output_result_to_csv(npu_dump_file.get(file_index),
                                                           golden_net_output_info.get(file_index),
                                                           compare_result, header)
                        logger.info(f"Compare Node_output:{file_index} completely.")
                    else:
                        logger.error(f"Failed to execute command: {' '.join(msaccucmp_cmd)}")
                        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DATA_ERROR)
                    file_index += 1
        return

    def save_net_output_result_to_csv(self, npu_file, golden_file, result, header):
        """
        save_net_output_result_to_csv
        """
        result_file_path = None
        result_file_backup_path = None
        npu_file_name = os.path.basename(npu_file)
        golden_file_name = os.path.basename(golden_file)
        for dir_path, _, files in os.walk(self.arguments.output_path):
            files = [file for file in files if file.endswith("csv")]
            if files:
                result_file_path = os.path.join(dir_path, files[0])
                result_file_backup = "{}_bak.csv".format(files[0].split(".")[0])
                result_file_backup_path = os.path.join(dir_path, result_file_backup)
                break
        try:
            # read result file and write it to backup file,update the result of compare Node_output
            Rule.input_file().check(result_file_path, will_raise=True)
            with ms_open(result_file_path, "r", max_size=TENSOR_MAX_SIZE) as fp_read:
                with ms_open(result_file_backup_path, 'w', newline="") as fp_write:
                    self._process_result_one_line(fp_write, fp_read, npu_file_name, golden_file_name, result)
            os.remove(result_file_path)
            os.rename(result_file_backup_path, result_file_path)
        except (OSError, SystemError, ValueError, TypeError, RuntimeError, MemoryError) as error:
            logger.error("Failed to write Net_output compare result")
            raise AccuracyCompareException(ACCURACY_COMPARISON_NET_OUTPUT_ERROR) from error
        finally:
            pass

    def execute_msaccucmp_command(self, cmd, catch=False):
        """
        Function Description:
            run the following command
        Parameter:
            cmd: command
        Return Value:
            status code
        """
        result = []
        header = []
        process = self.execute_command_line(cmd)
        while process.poll() is None:
            line = process.stdout.readline().strip()
            if line:
                compare_result, header_result = self._catch_compare_result(line, catch)
                result = compare_result if compare_result else result
                header = header_result if header_result else header
        return process.returncode, result, header

    def _check_msaccucmp_compare_support_args(self, compare_args):
        check_cmd = [self.python_version, self.msaccucmp_command_file_path, "compare", "-h"]
        check_cmd = filter_cmd(check_cmd)
        process = self.execute_command_line(check_cmd)
        while process.poll() is None:
            line = process.stdout.readline().strip()
            if line:
                line_decode = line.decode(encoding="utf-8")
                if compare_args in line_decode:
                    return True
        else:
            logger.warning(f'Current version does not support {compare_args} function')
            return False



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

from msprobe.core.common.file_utils import save_json


def create_kernel_config_json(dump_path, cur_rank):
    kernel_config_name = "kernel_config.json" if cur_rank == '' else f"kernel_config_{cur_rank}.json"
    kernel_config_path = os.path.join(dump_path, kernel_config_name)
    config_info = {
        "dump": {
            "dump_list": [],
            "dump_path": dump_path,
            "dump_mode": "all",
            "dump_op_switch": "on"
        }
    }
    save_json(kernel_config_path, config_info, indent=4)
    return kernel_config_path



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch
import numpy as np
from torch.nn import functional as F
from msprobe.core.common.log import logger

FLOAT_EPSILON = torch.finfo(torch.float).eps
NP_FLOAT_EPSILON = np.finfo(np.float32).eps
NAN = 'NaN'


def cosine_similarity(golden_data: torch.Tensor, my_data: torch.Tensor):
    if torch.all(golden_data == 0) and torch.all(my_data == 0):
        return 1.0, ''  # both are all 0, return similarity 1

    result = torch.cosine_similarity(golden_data.double(), my_data.double(), dim=0).item()  # Torch handle zero data
    return round(result, 10), ''  # Trunc to keeping only 10 decimals


def np_cosine_similarity(golden_data: np.ndarray, my_data: np.ndarray):
    if np.all(golden_data == 0) and np.all(my_data == 0):
        return 1.0, ''  # both are all 0, return similarity 1

    result = np.dot(golden_data, my_data) / (np.linalg.norm(golden_data) * np.linalg.norm(my_data))
    return round(result, 10), ''  #


def max_relative_error(golden_data: torch.Tensor, my_data: torch.Tensor):
    result = torch.where(
        torch.abs(golden_data) > FLOAT_EPSILON,
        torch.abs(my_data / golden_data - 1),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
        torch.tensor(0, dtype=golden_data.dtype),
    ).max()
    return result.item(), ''


def np_max_relative_error(golden_data: np.ndarray, my_data: np.ndarray):
    result = np.max(
        np.where(
            np.abs(golden_data) > NP_FLOAT_EPSILON,
            np.abs(my_data / golden_data - 1),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
            0
        )
    )
    return result, ''


def mean_relative_error(golden_data: torch.Tensor, my_data: torch.Tensor):
    result = torch.where(
        torch.abs(golden_data) > FLOAT_EPSILON,
        torch.abs(my_data / golden_data - 1),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
        torch.tensor(0, dtype=my_data.dtype),
    ).mean()
    return result.item(), ''


def np_mean_relative_error(golden_data: np.ndarray, my_data: np.ndarray):
    result = np.mean(
        np.where(
            np.abs(golden_data) > NP_FLOAT_EPSILON,
            np.abs(my_data / golden_data - 1),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
            0
        )
    )
    return result, ''


def max_absolute_error(golden_data: torch.Tensor, my_data: torch.Tensor):
    result = torch.where(
        torch.abs(golden_data) > FLOAT_EPSILON,
        torch.abs(my_data - golden_data),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
        torch.tensor(0, dtype=my_data.dtype),
    ).max()
    return result.item(), ''


def np_max_absolute_error(golden_data: np.ndarray, my_data: np.ndarray):
    result = np.max(
        np.where(
            np.abs(golden_data) > NP_FLOAT_EPSILON,
            np.abs(my_data - golden_data),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
            0
        )
    )
    return result, ''


def mean_absolute_error(golden_data: torch.Tensor, my_data: torch.Tensor):
    result = torch.where(
        torch.abs(golden_data) > FLOAT_EPSILON,
        torch.abs(my_data - golden_data),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
        torch.tensor(0, dtype=my_data.dtype),
    ).mean()
    return result.item(), ''


def np_mean_absolute_error(golden_data: np.ndarray, my_data: np.ndarray):
    result = np.mean(
        np.where(
            np.abs(golden_data) > NP_FLOAT_EPSILON,
            np.abs(my_data - golden_data),  # abs(aa - bb) / abs(bb) -> abs(aa / bb - 1)
            0
        )
    )
    return result, ''


def kl_divergence(golden_data: torch.Tensor, my_data: torch.Tensor):
    result = F.kl_div(F.log_softmax(my_data, dim=-1), F.softmax(golden_data, dim=-1), reduction="sum").item()
    return max(result, 0), ""


def np_kl_divergence(golden_data: np.ndarray, my_data: np.ndarray):
    golden_data = np.exp(golden_data - np.max(golden_data))
    my_data = np.exp(my_data - np.max(my_data))
    result = np.sum(golden_data * (np.log(golden_data) - my_data))
    return max(result, 0), ""


def relative_euclidean_distance(golden_data: torch.Tensor, my_data: torch.Tensor):
    ground_truth_square_num = (golden_data ** 2).sum()
    if ground_truth_square_num ** 0.5 <= FLOAT_EPSILON:
        return 0.0, ''

    result = ((my_data - golden_data) ** 2).sum() / ground_truth_square_num
    return torch.sqrt(result).item(), ''


def np_relative_euclidean_distance(golden_data: np.ndarray, my_data: np.ndarray):
    ground_truth_square_num = np.sum(golden_data ** 2)
    if np.sqrt(ground_truth_square_num) <= NP_FLOAT_EPSILON:
        return 0.0, ''

    result = np.sum((my_data - golden_data) ** 2) / ground_truth_square_num
    return np.sqrt(result), ''


def stat_absolute_error(golden_data, my_data):
    error = abs(my_data - golden_data)
    return error, ''


def stat_relative_error(golden_data, my_data):
    eps = np.finfo(np.float32).eps  
    denominator = max(abs(golden_data), eps)
    error = abs(my_data - golden_data) / denominator
    return error, ''


def l1_norm(golden_data: torch.Tensor, my_data: torch.Tensor):
    '''
    Function: compute l1_norm
    '''
    golden_data = golden_data.double()
    my_data = my_data.double()
    if golden_data.abs().sum() == 0 and my_data.abs().sum() == 0:
        return 0, ''
    elif golden_data.abs().sum() == 0:
        return 1, ''
    else:
        logger.debug(f"my data sum: {my_data.abs().sum()}")
        logger.debug(f"golden data sum: {golden_data.abs().sum()}")
        result = my_data.abs().sum() / golden_data.abs().sum() - 1
        return result.item(), ''


def register_custom_compare_algorithm(custom_compare_algorithm):
    import os
    import sys
    import importlib
    import inspect
    from msprobe.infer.utils.file_open_check import FileStat

    custom_compare_algorithm_split = custom_compare_algorithm.split(':')
    if len(custom_compare_algorithm_split) != 2:
        raise ValueError("custom_compare_algorithm should be in format '{python_file_path}:{function_name}'")
    file_path, func_name = custom_compare_algorithm_split
    file_path = os.path.expanduser(file_path)

    if not os.path.exists(file_path):
        raise ValueError(f"custom_compare_algorithm specified {file_path} not exists")
    if not file_path.endswith(".py"):
        raise ValueError("custom_compare_algorithm specified {file_path} is not a py file")
    
    file_stat = FileStat(file_path)
    if not file_stat.is_basically_legal('read', strict_permission=False):
        raise ValueError(f"custom_compare_algorithm specified {file_path} permission stat is illegal")

    file_dir, file_name = os.path.dirname(file_path), os.path.basename(file_path)
    if len(file_dir) > 0 and file_dir not in sys.path:
        sys.path.append(file_dir)

    custom_module_name = file_name.replace('.py', '')
    try:
        custom_module = importlib.import_module(custom_module_name)
    except Exception as ee:
        raise ValueError(f"import {custom_module_name} from {file_dir} failed") from ee

    custom_compare_func = getattr(custom_module, func_name, None)
    if custom_compare_func is None:
        raise ValueError(f"getting {func_name} from {custom_compare_algorithm} failed")
    if len(inspect.signature(custom_compare_func).parameters) != 2:
        raise ValueError(f"function {func_name} signature should have exact two parameters")

    try:
        ret = custom_compare_func(torch.ones([1]), torch.ones([1]))
    except Exception as e:
        raise ValueError(f"function {func_name} should recieve 2 torch tensor parameters") from e

    if not isinstance(ret, (list, tuple)) or len(ret) != 2:
        raise ValueError(f"function {func_name} should return 2 value in type ((float, int, str), str)")
    if not isinstance(ret[0], (float, int, str)) or not isinstance(ret[1], str):
        raise ValueError(f"function {func_name} should return 2 value in type ((float, int, str), str)")

    logger.info(f"Added custom comparing algorithm: {func_name}")
    CUSTOM_ALG_MAP[func_name] = custom_compare_func


CMP_ALG_MAP = {
    "cosine_similarity": cosine_similarity,
    "max_relative_error": max_relative_error,
    "mean_relative_error": mean_relative_error,
    "max_absolute_error": max_absolute_error,
    "mean_absolute_error": mean_absolute_error,
    "kl_divergence": kl_divergence,
    "relative_euclidean_distance": relative_euclidean_distance,
}


NP_CMP_ALG_MAP = {
    "cosine_similarity": np_cosine_similarity,
    "max_relative_error": np_max_relative_error,
    "mean_relative_error": np_mean_relative_error,
    "max_absolute_error": np_max_absolute_error,
    "mean_absolute_error": np_mean_absolute_error,
    "kl_divergence": np_kl_divergence,
    "relative_euclidean_distance": np_relative_euclidean_distance,
}

CUSTOM_ALG_MAP = {}


CMP_STATICTISC_MAP = {
    "stat_absolute_error": stat_absolute_error,
    "stat_relative_error": stat_relative_error,
}


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import re
import sqlite3
from typing import List, Tuple, Dict, Any
from functools import wraps

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import check_path_before_create, change_mode
from msprobe.core.common.const import FileCheckConst

SAFE_SQL_PATTERN = re.compile(r'^[a-zA-Z0-9_]+$')


def check_identifier_safety(name):
    """验证标识符是否安全（防止SQL注入）"""
    if not isinstance(name, str) or SAFE_SQL_PATTERN.match(name) is None:
        raise ValueError(f"Invalid SQL identifier: {name}, potential SQL injection risk!")


def _db_operation(func):
    """数据库操作装饰器，自动管理连接"""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        conn, curs = None, None
        try:
            conn, curs = self._get_connection()
            result = func(self, conn, curs, *args, **kwargs)
            return result  # 显式返回正常结果
            
        except sqlite3.Error as err:
            logger.error(f"Database operation failed: {err}")
            if conn:
                conn.rollback()
            return None  # 显式返回错误情况下的None
            
        finally:
            self._release_connection(conn, curs)
    return wrapper


class DBManager:
    """
    数据库管理类，封装常用数据库操作
    """

    DEFAULT_FETCH_SIZE = 10000
    DEFAULT_INSERT_SIZE = 10000
    MAX_ROW_COUNT = 100000000

    def __init__(self, db_path: str):
        """
        初始化DBManager
        :param db_path: 数据库文件路径
        :param table_config: 表配置对象
        """
        self.db_path = db_path

    @staticmethod
    def _get_where_sql(where_list):
        if not where_list:
            return "", tuple()

        where_clauses = []
        where_values = []
        if where_list:
            for col, val in where_list.items():
                check_identifier_safety(col)
                where_clauses.append(f"{col} = ?")
                where_values.append(val)
            if where_clauses:
                where_sql = " WHERE " + " AND ".join(where_clauses)
        return where_sql, tuple(where_values)

    @_db_operation
    def insert_data(self, conn: sqlite3.Connection, curs: sqlite3.Cursor,
                    table_name: str, data: List[Tuple], key_list: List[str] = None) -> int:
        """
        批量插入数据
        :param table_name: 表名
        :param data: 要插入的数据列表
        :param batch_size: 每批插入的大小
        :return: 插入的行数
        """
        check_identifier_safety(table_name)

        if not data:
            return 0
        columns = len(data[0])
        if key_list:
            if not isinstance(key_list, list):
                raise TypeError(
                    f"key_list must be a list, got {type(key_list)}"
                )
            if columns != len(key_list):
                raise ValueError(
                    f"When inserting into table {table_name}, the length of key list ({key_list})"
                    f"does not match the data({columns}).")
            for key in key_list:
                check_identifier_safety(key)

        batch_size = self.DEFAULT_INSERT_SIZE
        placeholders = ", ".join(["?"] * columns)
        if key_list:
            keys = ", ".join(key_list)
            sql = f"INSERT OR IGNORE INTO {table_name} ({keys}) VALUES ({placeholders})"
        else:
            sql = f"INSERT OR IGNORE INTO {table_name} VALUES ({placeholders})"

        inserted_rows = 0
        for i in range(0, len(data), batch_size):
            batch = data[i:i + batch_size]
            curs.executemany(sql, batch)
            inserted_rows += curs.rowcount

        conn.commit()
        return inserted_rows

    @_db_operation
    def select_data(self, conn: sqlite3.Connection, curs: sqlite3.Cursor,
                    table_name: str,
                    columns: List[str] = None,
                    where: dict = None) -> List[Dict]:
        """
        查询数据
        :param table_name: 表名
        :param columns: 要查询的列
        :param where: WHERE条件
        :return: 查询结果列表(字典形式)
        """
        check_identifier_safety(table_name)

        if not columns:
            raise ValueError("columns parameter cannot be empty, specify columns to select (e.g. ['id', 'name'])")
        if not isinstance(columns, list) or not all(isinstance(col, str) for col in columns):
            raise TypeError("columns must be a list of strings (e.g. ['id', 'name'])")
        
        for col in columns:
            check_identifier_safety(col)
        
        cols = ", ".join(columns)
        sql = f"SELECT {cols} FROM {table_name}"

        where_sql, where_parems = self._get_where_sql(where)
        curs.execute(sql + where_sql, where_parems)

        return [dict(row) for row in curs.fetchall()]

    @_db_operation
    def update_data(self, conn: sqlite3.Connection, curs: sqlite3.Cursor,
                    table_name: str, updates: Dict[str, Any],
                    where: dict = None) -> int:
        """
        更新数据
        :param table_name: 表名
        :param updates: 要更新的字段和值
        :param where: WHERE条件
        :param where_params: WHERE条件参数
        :return: 影响的行数
        """
        check_identifier_safety(table_name)
        if not updates:
            raise ValueError("columns parameter cannot be empty, specify it to update (e.g. {'name': 'xxx'}")
        if not isinstance(updates, dict):
            raise TypeError(f"updates must be a dictionary, got: {type(updates)}")
        for key in updates.keys():
            check_identifier_safety(key)

        set_clause = ", ".join([f"{k} = ?" for k in updates.keys()])
        sql = f"UPDATE {table_name} SET {set_clause}"

        params = tuple(updates.values())

        where_sql, where_parems = self._get_where_sql(where)

        curs.execute(sql + where_sql, params + where_parems)
        conn.commit()
        return curs.rowcount

    @_db_operation
    def execute_sql(self, conn: sqlite3.Connection, curs: sqlite3.Cursor,
                    sql: str, params: Tuple = None) -> List[Dict]:
        """
        执行自定义SQL查询
        :param sql: SQL语句
        :param params: 参数
        :return: 查询结果
        """
        curs.execute(sql, params or ())
        if sql.strip().upper().startswith("SELECT"):
            return [dict(row) for row in curs.fetchall()]
        conn.commit()
        return []

    def table_exists(self, table_name: str) -> bool:
        """
        :param table_name: 表名
        :return: 查询结果
        """
        result = self.select_data(
            table_name="sqlite_master",
            columns=["name"],
            where={"type": "table", "name": table_name}
        )
        return len(result) > 0

    @_db_operation
    def execute_multi_sql(self, conn: sqlite3.Connection, curs: sqlite3.Cursor,
                          sql_commands: List[str]) -> List[List[Dict]]:
        """
        批量执行多个SQL语句
        :param sql_commands: [sql1, sql2, ...]
        :return: 每个SELECT语句的结果列表
        """
        results = []
        for sql in sql_commands:
            curs.execute(sql)
            if sql.strip().upper().startswith("SELECT"):
                results.append([dict(row) for row in curs.fetchall()])
        conn.commit()
        return results

    def _get_connection(self) -> Tuple[sqlite3.Connection, sqlite3.Cursor]:
        """获取数据库连接和游标"""
        check_path_before_create(self.db_path)
        try:
            conn = sqlite3.connect(self.db_path)
            conn.row_factory = sqlite3.Row  # 使用Row工厂获取字典形式的结果
            curs = conn.cursor()
            return conn, curs
        except sqlite3.Error as err:
            logger.error(f"Database connection failed: {err}")
            raise

    def _release_connection(self, conn: sqlite3.Connection, curs: sqlite3.Cursor) -> None:
        """释放数据库连接"""
        try:
            if curs is not None:
                curs.close()
            if conn is not None:
                conn.close()
        except sqlite3.Error as err:
            logger.error(f"Failed to release database connection: {err}")
        change_mode(self.db_path, FileCheckConst.DATA_FILE_AUTHORITY)


class TrendSql:
    """趋势可视化数据库表参数类"""

    @staticmethod
    def create_monitoring_targets_table():
        """监测目标表"""
        return """
        CREATE TABLE IF NOT EXISTS monitoring_targets (
            target_id INTEGER PRIMARY KEY AUTOINCREMENT,
            target_name TEXT NOT NULL,
            vpp_stage INTEGER NOT NULL,
            micro_step INTEGER NOT NULL DEFAULT 0,
            UNIQUE(target_name, vpp_stage, micro_step) 
        )"""

    @staticmethod
    def create_monitoring_metrics_table():
        """监测指标表"""
        return """
        CREATE TABLE IF NOT EXISTS monitoring_metrics (
            metric_id INTEGER PRIMARY KEY AUTOINCREMENT,
            metric_name TEXT UNIQUE NOT NULL
        )"""

    @staticmethod
    def get_metric_mapping_sql():
        """从monitoring_metrics表获取所有metric名称和ID"""
        return """
        SELECT metric_id, metric_name
        FROM monitoring_metrics
        """

    @staticmethod
    def get_global_stats_sql():
        """从global_stats表获取最新的统计配置"""
        return """
        SELECT * FROM global_stats 
        ORDER BY ROWID DESC 
        LIMIT 1
        """

    @staticmethod
    def create_global_stats_table(columns_config):
        """根据字典配置创建全局统计表"""
        # 根据字典的键值类型动态创建列
        column_definitions = []
        for column_name, column_value in columns_config.items():
            if isinstance(column_value, (list, set)):
                column_definitions.append(f"{column_name} TEXT DEFAULT NULL")
            elif isinstance(column_value, (int, float)):
                column_definitions.append(f"{column_name} INTEGER DEFAULT 0")
        
        create_sql = f"""
        CREATE TABLE IF NOT EXISTS global_stats (
            {', '.join(column_definitions)}
        )"""
        return create_sql

    @staticmethod
    def create_tags_table():
        # 新增：标签表
        return """
        CREATE TABLE IF NOT EXISTS monitoring_tags (
            tag_id INTEGER PRIMARY KEY AUTOINCREMENT,
            tag_name TEXT NOT NULL,
            category TEXT NOT NULL,
            metric_id INTEGER NOT NULL,
            UNIQUE(tag_name, category, metric_id),
            FOREIGN KEY (metric_id) REFERENCES monitoring_metrics(metric_id)
        )"""

    @staticmethod
    def create_tag_mapping_table():
        return """
        CREATE TABLE IF NOT EXISTS tag_target_mapping (
            tag_id INTEGER NOT NULL,
            target_id INTEGER NOT NULL,
            PRIMARY KEY (tag_id, target_id),
            FOREIGN KEY (tag_id) REFERENCES monitoring_tags(tag_id),
            FOREIGN KEY (target_id) REFERENCES monitoring_target(target_id)
        )"""

    @staticmethod
    def create_trend_table(stats):
        stat_columns = [f"{stat} REAL DEFAULT NULL" for stat in stats]
        create_sql = f"""
        CREATE TABLE IF NOT EXISTS trend_data (
            rank INTEGER NOT NULL,
            step INTEGER NOT NULL,
            target_id INTEGER NOT NULL,
            metric_id INTEGER NOT NULL,
            {', '.join(stat_columns)},
            PRIMARY KEY (rank, step, target_id, metric_id),
            FOREIGN KEY (target_id) REFERENCES monitoring_targets(target_id),
            FOREIGN KEY (metric_id) REFERENCES monitoring_metrics(metric_id)
        ) WITHOUT ROWID"""
        return create_sql
    
    @classmethod
    def get_table_definition(cls, table_name=""):
        """
        获取表定义SQL
        :param table_name: 表名
        :return: 建表SQL语句
        :raises ValueError: 当表名不存在时
        """
        table_creators = {
            "monitoring_targets": cls.create_monitoring_targets_table,
            "monitoring_metrics": cls.create_monitoring_metrics_table,
            "monitoring_tags": cls.create_tags_table,
            "tag_target_mapping": cls.create_tag_mapping_table,
        }
        if not table_name:
            return [table_creators.get(table, lambda x: "")() for table in table_creators]
        if table_name not in table_creators:
            raise ValueError(f"Unsupported table name: {table_name}")
        return table_creators[table_name]()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from collections import OrderedDict
from collections.abc import Iterable
from typing import Dict, List, Optional, Set, Tuple

from msprobe.core.common.const import MonitorConst, Const, Data2DBConst
from msprobe.core.common.db_manager import DBManager, TrendSql


def update_ordered_dict(main_dict: OrderedDict, new_list: List) -> OrderedDict:
    """Update ordered dictionary with new items"""
    for item in new_list:
        if item not in main_dict:
            main_dict[item] = None
    return main_dict


def get_ordered_list(data: Iterable, order: Iterable) -> List[str]:
    if not isinstance(data, Iterable) or not isinstance(order, Iterable):
        return []
    return [value for value in order if value in data]

def get_tags_from_target(target_name):
    """
    full target name to tags
    
    :param target_name: e.g. layers.22.self_attention.linear_qkv.layer_norm_weight
    :return tags: set of tags
    """
    tags = set()
    parts = target_name.split(Const.SEP)[::-1]
    max_idx = len(parts) - 1
    for i, part in enumerate(parts):
        if not part.isdigit():
            tags.add((part, Data2DBConst.TAG_MODULE))
            continue
        if i == max_idx or parts[i+1].isdigit():  # 确保数字前面有非数字部分, 组成layer标签
            continue
        layer_tag = (f"{parts[i+1]}.{parts[i]}", Data2DBConst.TAG_LAYER)
        tags.add(layer_tag)
    return tags
    

class MonitorDB:
    """Main class for monitoring database operations"""

    def __init__(self, db_path: str):
        self.db_path = db_path
        self.db_manager = DBManager(db_path)

    def init_schema(self) -> None:
        """Initialize database schema"""
        self.db_manager.execute_multi_sql(TrendSql.get_table_definition())

    def insert_dimensions(
        self,
        targets: Dict[str,OrderedDict],
    ) -> None:
        """Insert dimension data into database"""
        # Insert targets
        metrics = get_ordered_list(set(targets.keys()), MonitorConst.METRICS_TRENDVIS_SUPPORTED)
        for metric_name in metrics:
            self.db_manager.insert_data(
                "monitoring_targets",
                [(name, vpp_stage, micro_step)
                for (name, vpp_stage, micro_step) in targets[metric_name]],
                key_list=["target_name", "vpp_stage", "micro_step"]
            )

        # Insert metrics
        self.db_manager.insert_data(
            "monitoring_metrics",
            [(metric,) for metric in metrics],
            key_list=["metric_name"]
        )

    def insert_rows(self, rows):
        table_name = "trend_data"
        inserted = self.db_manager.insert_data(table_name, rows)
        inserted = 0 if inserted is None else inserted
        return inserted

    def init_global_stats_data(self, config: Dict):
        """初始化全局统计表数据"""
        # 准备插入数据
        self.db_manager.execute_sql(
            TrendSql.create_global_stats_table(config)
        )
        
        column_values = []
        for _, column_value in config.items():
            if isinstance(column_value, (list, set)):
                column_values.append(",".join(column_value))
            elif isinstance(column_value, (int, float)):
                column_values.append(column_value)
        self.db_manager.insert_data("global_stats", [column_values,])

    def get_metric_mapping(self) -> Dict[str, Tuple[int, List[str]]]:
        """获取metric名称到ID的映射及对应的统计信息"""
        metric_results = self.db_manager.execute_sql(
            TrendSql.get_metric_mapping_sql()
        )        
        global_stats_result = self.db_manager.execute_sql(
            TrendSql.get_global_stats_sql()
        )
        if not global_stats_result:
            return {}
        
        global_stats_config = global_stats_result[0]
        # 构建映射字典
        metric_mapping = {}
        for row in metric_results:
            metric_name = row["metric_name"]
            metric_id = row["metric_id"]
            
            # 从global_stats中获取该metric对应的统计信息
            stats_value = global_stats_config.get(metric_name)
            ordered_stats = []
            if stats_value:
                stats_list = stats_value.split(",") if isinstance(stats_value, str) else []
                ordered_stats = get_ordered_list(stats_list, MonitorConst.OP_MONVIS_SUPPORTED)                
            metric_mapping[metric_name] = (metric_id, ordered_stats)
        return metric_mapping

    def get_target_mapping(self) -> Dict[Tuple[str, int, int], int]:
        """Get target mapping dictionary"""
        results = self.db_manager.select_data(
            table_name="monitoring_targets",
            columns=["target_id", "target_name", "vpp_stage", "micro_step"]
        )
        if not results:
            return {}
        return {
            (row["target_name"], row["vpp_stage"], row["micro_step"]): row["target_id"]
            for row in results
        }

    def create_trend_data(self, stats):
        self.db_manager.execute_sql(TrendSql.create_trend_table(stats))

    def _get_metric_id(self, metric_name: str) -> Optional[int]:
        """Get metric ID by name"""
        result = self.db_manager.select_data(
            table_name="monitoring_metrics",
            columns=["metric_id"],
            where={"metric_name": metric_name}
        )
        return result[0]["metric_id"] if result else None

    def extract_tags_from_processed_targets(self, targets:dict , metric_id_dict: dict, target_dict: dict):
        """从已处理的targets中提取标签并建立映射关系"""

        for metric_name in targets:
            tag_info_to_id = {}
            mapping_data = []
            new_tags_data = set()
            target_to_tags = {}
            metric_id = metric_id_dict.get(metric_name, [None,])[0]
            for target in targets[metric_name]:
                target_id = target_dict[target]
                name, _, _ = target
                tags = get_tags_from_target(name)
                new_tags_data = new_tags_data.union(tags)
                target_to_tags[target_id] = tags

            # 插入新标签
            if not new_tags_data:
                continue
            self.db_manager.insert_data(
                "monitoring_tags",
                [(tag_name, category, metric_id)
                    for tag_name, category in new_tags_data],
                key_list=["tag_name", "category", "metric_id"]
            )

            # 重新获取tag_id用于映射
            tag_records = self.db_manager.select_data(
                "monitoring_tags",
                columns=["tag_id", "tag_name", "category", "metric_id"]
            )
            for row in tag_records:
                tag_info_to_id[(row["tag_name"], row["category"],
                                row["metric_id"])] = row["tag_id"]

            for target_id, tags_list in target_to_tags.items():
                for tag_name, category in tags_list:
                    tag_id = tag_info_to_id.get(
                        (tag_name, category, metric_id))
                    if tag_id:
                        mapping_data.append((target_id, tag_id))

            # 插入映射关系
            if mapping_data:
                self.db_manager.insert_data(
                    "tag_target_mapping",
                    mapping_data,
                    key_list=["target_id", "tag_id"]
                )


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from typing import Union

from msprobe.infer.utils.check.checker import Checker, CheckResult, rule


class ListChecker(Checker):

    @rule()
    def is_list(self) -> Union["ListChecker", CheckResult]:
        return isinstance(self.instance, list), "Input object is not a list."

    @rule()
    def is_list_not_empty(self) -> Union["ListChecker", CheckResult]:
        return bool(self.instance), "List is empty."

    @rule()
    def is_element_valid(self, check_rule) -> Union["ListChecker", CheckResult]:
        is_pass = self.is_list().passed
        if not is_pass:
            return is_pass, "Input object is not a list."
        err_msgs = []
        for element in self.instance:
            rule_is_pass = check_rule.check(element)
            if not rule_is_pass:
                is_pass = False
                err_msgs.append(f"{element}: {str(rule_is_pass)}")
        return is_pass, "\n".join(err_msgs)

    @rule()
    def is_length_valid(self, min_length=None, max_length=None) -> Union["ListChecker", CheckResult]:
        length = len(self.instance)
        length_valid = True
        err_msg = []
        if min_length is not None and length < min_length:
            length_valid = False
            err_msg.append(f"List is shorter than the minimum length of {min_length}.")
        if max_length is not None and length > max_length:
            length_valid = False
            err_msg.append(f"List is longer than the maximum length of {max_length}.")
        return length_valid, "\n".join(err_msg)



# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os

CANN_PATH = os.environ.get('ASCEND_TOOLKIT_HOME', "/usr/local/Ascend/ascend-toolkit/latest")


class DumpArgsAdapter:
    def __init__(
            self,
            model_path,
            input_data="",
            cann_path=CANN_PATH,
            output_path="./output",
            input_shape="",
            rank="0",
            dym_shape_range="",
            onnx_fusion_switch=True,
            custom_op="",
            dump=True,
            single_op="",
            output_size=""
    ):
        self.golden_path = model_path
        self.input_data = input_data
        self.cann_path = cann_path
        self.output_path = output_path
        self.input_shape = input_shape
        self.rank = rank
        self.dym_shape_range = dym_shape_range
        self.onnx_fusion_switch = onnx_fusion_switch
        self.custom_op = custom_op
        self.dump = dump
        self.single_op = single_op
        self.output_size = output_size



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from __future__ import annotations

from typing import Any, Dict, Optional

from msprobe.core.common.framework_adapter import FmkAdp
from msprobe.core.common.file_utils import load_json
from msprobe.core.common.log import logger
from msprobe.core.monitor.utils import validate_int_arg, validate_step_count_per_record, validate_ranks
from msprobe.core.monitor_v2.base import BaseMonitorV2
from msprobe.core.monitor_v2.factory import MonitorFactory
from msprobe.core.monitor_v2.writer import CSVWriterV2


_FRAMEWORK_ALIASES = {
    "pytorch": "pytorch",
    "pt": "pytorch",
    "torch": "pytorch",
    "mindspore": "mindspore",
    "ms": "mindspore",
}


def _resolve_framework(config: Dict[str, Any], fr: Optional[str]) -> str:
    candidate = fr or config.get("framework", "pytorch")
    key = str(candidate).strip().lower()
    if key not in _FRAMEWORK_ALIASES:
        raise ValueError(f"[monitor_v2] Unsupported framework '{candidate}'")
    return _FRAMEWORK_ALIASES[key]


class TrainerMonitorV2:
    """
    Thin orchestrator for monitor_v2.

    Usage (PyTorch):
        mon = TrainerMonitorV2("config.json", fr="pytorch")
        mon.start(model=model)
        for step in range(num_steps):
            ...
            mon.step()
        mon.stop()

    Config (minimal example):
        {
          "framework": "pytorch",
          "output_dir": "./monitor_v2_out",
          "rank": [0, 1],                # optional, list of ranks to monitor
          "start_step": 0,               # optional, first step to write
          "step_interval": 1,            # optional, write every N steps
          "step_count_per_record": 1,    # optional, steps per output file
          "monitors": {
            "module": {
              "enabled": true,
              "targets": [],
              "ops": ["min", "max", "mean", "norm", "nans"]
            }
          }
        }
    """

    def __init__(self, config_path: str, fr: Optional[str] = None) -> None:
        self.config_path = config_path
        self.config: Dict[str, Any] = load_json(config_path)
        if not isinstance(self.config, dict):
            raise TypeError("[monitor_v2] Configuration must be a dictionary.")

        self.framework = _resolve_framework(self.config, fr)

        # Ensure framework adapter is initialised so that rank resolution and
        # tensor ops work correctly for the selected framework.
        try:
            FmkAdp.set_fmk(self.framework)
        except Exception as exc:
            # Best-effort; fall back to default adapter behaviour.
            logger.warning(f"[monitor_v2] Failed to set framework adapter to '{self.framework}': {exc}")

        # Current rank is always resolved from framework adapter.
        self.rank = FmkAdp.get_rank_id()

        # `rank` in config is a list (or single int) of ranks to monitor.
        self._target_ranks = self._parse_target_ranks(self.config.get("rank", []))

        # Step control: start_step / stop_step define active window,
        # step_interval controls how often we write.
        # step_count_per_record controls how many steps go into one output file.
        self.start_step = validate_int_arg(self.config.get("start_step"), "start_step", 0, 0)
        self.step_interval = validate_int_arg(self.config.get("step_interval"), "step_interval", 1, 1)
        try:
            validate_step_count_per_record(self.config.get("step_count_per_record", 1))
            self.step_count_per_record = int(self.config.get("step_count_per_record", 1))
        except Exception as exc:
            logger.warning(f"[monitor_v2] Validate step_count_per_record failed: {exc}; fallback to 1.")
            self.step_count_per_record = 1
        # collect_times: 最大采集次数，达到后停止写出；默认大值表示“几乎一直采集”
        # stop_step：优先使用显式配置，其次由 collect_times 推导，最后使用一个极大默认值
        self.collect_times = validate_int_arg(self.config.get("collect_times"), "collect_times", 1, 100_000_000)
        explicit_stop = validate_int_arg(self.config.get("stop_step"), "stop_step", self.start_step, None)
        if explicit_stop is not None:
            self.stop_step = int(explicit_stop)
        else:
            # start_step + collect_times * interval 对齐“采集次数”语义
            self.stop_step = self.start_step + self.collect_times * max(self.step_interval, 1)
        self.current_step = 0
        self._collected_steps = 0
        self.output_dir = self.config.get("output_dir", "./")

        self.monitors = self._build_monitors()
        fmt = str(self.config.get("format", "csv")).strip().lower()
        if fmt != "csv":
            raise ValueError(f"[monitor_v2] Unsupported format '{fmt}', only 'csv' is currently supported.")
        self.writer = CSVWriterV2(
            self.output_dir,
            rank=self.rank,
            async_write=bool(self.config.get("async_write", False)),
        )
        self._writer_closed = False
        self._optimizer = None
        self._orig_step = None
        self._step_patched = False
        self._patch_optimizer_step = bool(self.config.get("patch_optimizer_step", False))
        self._warned_manual_step_with_optimizer = False

    def start(self, model: Any = None, optimizer: Any = None, **context: Any) -> None:
        """
        Start all configured monitors.
        """
        if not self._is_target_rank():
            return
        static_ctx = {
            "rank": self.rank,
            "output_dir": self.output_dir,
            "step_provider": lambda: self.current_step,
        }
        static_ctx.update(context)
        for mon in self.monitors:
            mon.start(model=model, optimizer=optimizer, **static_ctx)
        if optimizer is not None:
            self._optimizer = optimizer
            if "patch_optimizer_step" not in self.config:
                self._patch_optimizer_step = True
        if self._optimizer is not None and self._patch_optimizer_step:
            self._patch_step_via_optimizer()

    def step(self, from_optimizer: bool = False) -> None:
        """
        Increment step counter and collect/write data from monitors.
        """
        if not self._is_target_rank():
            return
        if self._patch_optimizer_step and self._optimizer is not None and not from_optimizer:
            if not self._warned_manual_step_with_optimizer:
                logger.warning(
                    "[monitor_v2] step() called manually while optimizer patch is enabled; "
                    "skip to avoid double increment. Set patch_optimizer_step=false to "
                    "manage steps manually."
                )
                self._warned_manual_step_with_optimizer = True
            return
        step_val = self.current_step
        self.current_step += 1

        # step_interval 语义：每隔多少个 step 写一次。
        # step_count_per_record 语义：每多少个 step 写到一个 csv 文件。
        if not self._should_collect_step(step_val):
            for mon in self.monitors:
                collect_fn = getattr(mon, "collect", None)
                if not callable(collect_fn):
                    continue
                try:
                    collect_fn()
                except Exception as exc:
                    # 收集失败不影响训练流程，直接跳过该 monitor
                    logger.warning(f"[monitor_v2] Failed to collect data '{mon.__class__.__name__}': {exc}")
            return

        for mon in self.monitors:
            collect_fn = getattr(mon, "collect", None)
            if not callable(collect_fn):
                continue
            out = collect_fn()
            if not out:
                continue
            # Normalise monitor payload so that individual monitors
            # only need to care about producing "rows". Trainer is
            # responsible for adding common metadata.
            if not isinstance(out, dict):
                continue
            if not out.get("rows"):
                continue
            if "slug" not in out:
                slug = getattr(mon, "slug", None)
                if not isinstance(slug, str):
                    slug = mon.__class__.__name__.lower()
                out["slug"] = slug
            if "monitor" not in out:
                out["monitor"] = out.get("slug")
            if "rank" not in out:
                out["rank"] = self.rank
            if "step" not in out:
                out["step"] = step_val
            if "step_interval" not in out:
                out["step_interval"] = self.step_interval
            if "step_count_per_record" not in out:
                out["step_count_per_record"] = self.step_count_per_record
            if "start_step" not in out:
                out["start_step"] = self.start_step
            self.writer.write_monitor_data(out)
        # 仅在实际写出数据时增加采集次数
        self._collected_steps += 1

    def stop(self) -> None:
        """
        Stop all monitors and close writer.
        """
        if not self._is_target_rank():
            return
        for mon in self.monitors:
            mon.stop()
        self._restore_optimizer_step()
        self._close_writer()

    def _build_monitors(self):
        monitors_cfg = self.config.get("monitors", {})
        if not isinstance(monitors_cfg, dict):
            logger.warning("[monitor_v2] Monitors config should be a dict; using empty config.")
            monitors_cfg = {}
        monitors = []

        for name, cfg in monitors_cfg.items():
            if not isinstance(cfg, dict) or not cfg.get("enabled", name == "module"):
                continue
            mon = MonitorFactory.create(self.framework, name)
            if mon is None:
                continue
            mon_cfg = {k: v for k, v in cfg.items() if k != "enabled"}
            setattr(mon, "slug", name)
            mon.set_config(mon_cfg)
            monitors.append(mon)

        return monitors

    def _parse_target_ranks(self, rank_cfg: Any) -> set[int]:
        if isinstance(rank_cfg, int):
            return {rank_cfg}
        if isinstance(rank_cfg, list):
            try:
                validate_ranks(rank_cfg)
                return {int(r) for r in rank_cfg}
            except Exception as exc:
                logger.warning(f"[monitor_v2] Validate rank list failed: {exc}; fallback to all ranks.")
                return set()
        logger.warning("[monitor_v2] Rank config should be int or list; fallback to all ranks.")
        return set()

    def _should_collect_step(self, step: int) -> bool:
        """
        Global step filter for monitor_v2:
        - Only write when start_step <= step < stop_step
        - And (step - start_step) % step_interval == 0 when interval > 1
        - Respect collect_times: stop after N successful writes
        """
        if self.collect_times > 0 and self._collected_steps >= self.collect_times:
            return False
        if not (self.start_step <= step < self.stop_step):
            return False
        if self.step_interval <= 1:
            return True
        return (step - self.start_step) % self.step_interval == 0

    def _is_target_rank(self) -> bool:
        """
        Whether current rank should participate in monitor_v2 collection.
        Empty `rank` list in config means all ranks.
        """
        return not self._target_ranks or self.rank in self._target_ranks

    def _close_writer(self) -> None:
        if self._writer_closed:
            return
        try:
            self.writer.close()
        finally:
            self._writer_closed = True

    def _patch_step_via_optimizer(self) -> None:
        if self._step_patched or self._optimizer is None:
            return
        step_fn = getattr(self._optimizer, "step", None)
        if not callable(step_fn):
            return
        self._orig_step = step_fn

        def wrapped_step(*args: Any, **kwargs: Any):
            out = self._orig_step(*args, **kwargs)
            try:
                self.step(from_optimizer=True)
            except Exception as exc:
                logger.warning(f"[monitor_v2] Failed to run step from optimizer patch: {exc}")
            return out

        self._optimizer.step = wrapped_step  # type: ignore[assignment]
        self._step_patched = True

    def _restore_optimizer_step(self) -> None:
        if not self._step_patched or self._optimizer is None:
            return
        if self._orig_step is not None:
            self._optimizer.step = self._orig_step  # type: ignore[assignment]
        self._step_patched = False
        self._orig_step = None



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import functools
from collections import defaultdict

import torch
import torch.nn as nn
import torch.utils.hooks as full_hooks

from msprobe.pytorch.common.utils import register_forward_pre_hook, is_float8_tensor


class HOOKModule(nn.Module):
    module_count = defaultdict(int)

    def __init__(self, hook_build_func) -> None:
        super(HOOKModule, self).__init__()
        prefix = self.prefix_api_name if hasattr(self, "prefix_api_name") else ""
        op_is_distributed = self.op_is_distributed if hasattr(self, "op_is_distributed") else False
        if callable(hook_build_func):
            hook_set = hook_build_func(prefix)
            register_forward_pre_hook(self, hook_set.forward_pre_hook)
            if op_is_distributed:
                self.distributed_forward_hook = hook_set.distributed_forward_hook

    def __call__(self, *args, **kwargs):
        return self._call_func(*args, **kwargs)

    @staticmethod
    def reset_module_stats():
        HOOKModule.module_count = defaultdict(int)

    @staticmethod
    def add_module_count(name):
        HOOKModule.module_count[name] += 1

    @staticmethod
    def get_module_count(name):
        return HOOKModule.module_count[name]

    def _call_func(self, *args, **kwargs):
        full_backward_hooks, non_full_backward_hooks = [], []
        if len(self._backward_hooks) > 0:
            full_backward_hooks, non_full_backward_hooks = self._get_backward_hooks()
        for hook in self._forward_pre_hooks.values():
            hook(self, args, kwargs)
        bw_hook = None
        if len(full_backward_hooks) > 0:
            bw_hook = full_hooks.BackwardHook(self, full_backward_hooks)
            args = bw_hook.setup_input_hook(args)
        if torch._C._get_tracing_state():
            result = self._slow_forward(*args, **kwargs)
        else:
            result = self.forward(*args, **kwargs)
        for hook in self._forward_hooks.values():
            hook_result = hook(self, args, kwargs, result)
            if hook_result is not None:
                result = hook_result
        if bw_hook:
            result = bw_hook.setup_output_hook(result)
        if len(non_full_backward_hooks) > 0:
            var = result
            while not isinstance(var, torch.Tensor):
                if isinstance(var, dict):
                    var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
                elif isinstance(var, (list, tuple)):
                    if var:
                        var = var[0]
                    else:
                        return result
                else:
                    return result

            if is_float8_tensor(var) or not (var.requires_grad and torch.is_grad_enabled()):
                return result

            grad_fn = var.grad_fn
            if grad_fn is not None:
                for hook in non_full_backward_hooks:
                    wrapper = functools.partial(hook, self)
                    functools.update_wrapper(wrapper, hook)
                    grad_fn.register_hook(wrapper)
                self._maybe_warn_non_full_backward_hook(args, result, grad_fn)
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import argparse
import os
import sys

try:
    import torch_npu
except ImportError:
    is_gpu = True
    current_device = "cuda"
else:
    is_gpu = False
    current_device = "npu"
import torch
from tqdm import tqdm
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check import generate_device_params, get_api_info
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import exec_api, is_unsupported_api, ExecParams
from msprobe.core.common.file_utils import check_link, FileChecker
from msprobe.pytorch.api_accuracy_checker.common.utils import extract_basic_api_segments
from msprobe.core.common.const import FileCheckConst, Const
from msprobe.core.common.utils import check_op_str_pattern_valid
from msprobe.pytorch.common.log import logger
from msprobe.pytorch.common.parse_json import parse_json_info_forward_backward
from msprobe.core.common.decorator import recursion_depth_decorator


def check_tensor_overflow(x):
    if isinstance(x, torch.Tensor) and x.numel() != 0 and x.dtype != torch.bool:
        if len(x.shape) == 0:
            tensor_max = x.cpu().detach().float().numpy().tolist()
            tensor_min = tensor_max
        else:
            tensor_max = torch.max(x).cpu().detach().float().numpy().tolist()
            tensor_min = torch.min(x).cpu().detach().float().numpy().tolist()
        # inf
        if tensor_max == float('inf') or tensor_min == float('-inf'):
            return True
        # nan
        elif tensor_max != tensor_max or tensor_min != tensor_min:
            return True
        else:
            return False
    elif isinstance(x, bool) or isinstance(x, int) or isinstance(x, float):
        if x == float('inf') or x == float('-inf') or x != x:
            return True
        else:
            return False
    else:
        return False


@recursion_depth_decorator("check_data_overflow")
def check_data_overflow(x, device):
    if isinstance(x, (tuple, list)):
        if not x:
            return False
        return any(check_data_overflow(item, device) for item in x)
    else:
        if device == Const.CPU_LOWERCASE:
            return check_tensor_overflow(x)
        else:
            return torch_npu.npu.utils.npu_check_overflow(x)


@recursion_depth_decorator("is_bool_output")
def is_bool_output(x):
    if isinstance(x, (tuple, list)):
        if not x:
            return False
        return any(is_bool_output(item) for item in x)
    else:
        return isinstance(x, bool)


def run_overflow_check(forward_file):
    logger.info("start acc_check test")
    forward_content, _, real_data_path = parse_json_info_forward_backward(forward_file)
    if real_data_path:
        dump_path = os.path.dirname(forward_file)
        real_data_path = os.path.join(dump_path, Const.DUMP_TENSOR_DATA)
    for api_full_name, api_info_dict in tqdm(forward_content.items()):
        check_op_str_pattern_valid(api_full_name)
        if is_unsupported_api(api_full_name, is_overflow_check=True):
            continue
        try:
            run_torch_api(api_full_name, api_info_dict, real_data_path)
        except Exception as err:
            _, api_name, _ = api_full_name.split(Const.SEP)
            if "not implemented for 'Half'" in str(err):
                logger.warning(f"API {api_name} not support half tensor in CPU. This API does not support overflow "
                               "check, so it will be skipped.")
            elif "expected scalar type Long" in str(err):
                logger.warning(f"API {api_name} not support int32 tensor in CPU, please add {api_name} to CONVERT_API "
                               "'int32_to_int64' list in accuracy_tools/msprobe/core/common/const.py file.")
            elif "could not create a primitive descriptor for a matmul primitive" in str(err):
                logger.warning(f"API {api_name} not support matmul primitive in CPU due to pytorch bug, "
                               "so it will be skipped.")
            else:
                logger.error(f"Run {api_full_name} acc_check Error: %s" % str(err))


def run_torch_api(api_full_name, api_info_dict, real_data_path):
    torch.npu.clear_npu_overflow_flag()
    api_type, api_name = extract_basic_api_segments(api_full_name)
    args, kwargs, need_grad = get_api_info(api_info_dict, api_name, real_data_path)
    if not need_grad:
        logger.warning("%s function with out=... arguments don't support automatic differentiation, skip backward." 
                       % api_full_name)
    device_info_kwargs = kwargs.get(Const.DEVICE)
    if device_info_kwargs and device_info_kwargs.get(Const.VALUE):
        kwargs[Const.DEVICE] = current_device
    npu_args, npu_kwargs, _ = generate_device_params(args, kwargs, False, api_name)
    if kwargs.get(Const.DEVICE):
        del kwargs[Const.DEVICE]
    cpu_exec_params = ExecParams(api_type, api_name, Const.CPU_LOWERCASE, args, kwargs, False, None)
    device_exec_params = ExecParams(api_type, api_name, Const.NPU_LOWERCASE, npu_args, npu_kwargs, False, None)
    out = exec_api(cpu_exec_params)
    npu_out = exec_api(device_exec_params)
    if out is None and npu_out is None:
        logger.warning("The %s overflow is a normal overflow, out and npu_out is None." % api_full_name)
        return
    if is_bool_output(out) or is_bool_output(npu_out):
        logger.warning("The output of %s is bool type.This dtype not support overflow, so it will be skipped."
                       % api_full_name)
        return

    cpu_overflow = check_data_overflow(out, Const.CPU_LOWERCASE)
    npu_overflow = check_data_overflow(npu_out, Const.NPU_LOWERCASE)
    if cpu_overflow == npu_overflow:
        logger.warning("The %s overflow is a normal overflow." % api_full_name)
    else:
        logger.warning("The %s overflow is an abnormal overflow." % api_full_name)
    return


def _run_overflow_check_parser(parser):
    parser.add_argument("-api_info", "--api_info_file", dest="api_info_file", default="",
                        help="<Required> The api param tool result file: generate from api param tool, "
                             "a json file.",
                        required=True)
    parser.add_argument("-j", "--jit_compile", dest="jit_compile", help="<optional> whether to turn on jit compile",
                        default=False, required=False)
    parser.add_argument("-d", "--device", dest="device_id", type=int, help="<optional> set NPU device id to acc_check",
                        default=0, required=False)


def _run_overflow_check(parser=None):
    if not parser:
        parser = argparse.ArgumentParser()
    _run_overflow_check_parser(parser)
    args = parser.parse_args(sys.argv[1:])
    _run_overflow_check_command(args)
    logger.info("acc_check task completed.")


def _run_overflow_check_command(args):
    torch.npu.set_compile_mode(jit_compile=args.jit_compile)
    npu_device = "npu:" + str(args.device_id)
    api_info_file_checker = FileChecker(file_path=args.api_info_file, path_type=FileCheckConst.FILE, 
                                            ability=FileCheckConst.READ_ABLE, file_type=FileCheckConst.JSON_SUFFIX)
    api_info = api_info_file_checker.common_check()
    try:
        torch.npu.set_device(npu_device)
    except Exception as error:
        logger.error(f"Set NPU device id failed. device id is: {args.device_id}")
        raise NotImplementedError from error
    run_overflow_check(api_info)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import inspect
import os
import re
from dataclasses import dataclass

import numpy as np

from mindspore import nn, Tensor, ops, _no_grad
from mindspore import communication
from mindspore.communication import comm_func, get_rank

from msprobe.core.common.const import MonitorConst, Const
from msprobe.core.common.file_utils import load_yaml
from msprobe.mindspore.monitor.utils import get_metrics, get_summary_writer_tag_name

enable_communication = True
try:
    from mindspore.communication.comm_func import CommHandle as CommHandle_
except ImportError:
    enable_communication = False


RANK = None

OpsPath = os.path.join(os.path.dirname(__file__), "distributed_ops.yaml")
WrapDistributedOps = load_yaml(OpsPath).get("communication.comm_func", [])

StackBlackListPath = os.path.join(os.path.dirname(__file__), "stack_blacklist.yaml")
StackBlackList = load_yaml(StackBlackListPath).get("stack", [])


@dataclass(frozen=True)
class CatchDataParams:
    cc_context: object
    cc_name: str
    ops_list: list
    args: tuple
    prefix: str
    call_idx: int = 0

distributed_func = {}
for f in dir(comm_func):
    distributed_func[f] = getattr(comm_func, f)

ORIGIN_WAIT = CommHandle_.wait if enable_communication else None
PENDING_ASYNC_CC_BY_HANDLE = {}


def get_distributed_ops():
    global WrapDistributedOps
    _all_distributed_ops = dir(comm_func)
    return set(WrapDistributedOps) & set(_all_distributed_ops)


class DistributedOPTemplate(nn.Cell):
    def __init__(self, op_name, pre_hooks, post_hooks):
        super(DistributedOPTemplate, self).__init__()
        self.op_name_ = str(op_name)
        self.__name__ = self.op_name_
        self.idx = 0
        self.cc_hooks = []
        for pre_hook in pre_hooks:
            handle = self.register_forward_pre_hook(pre_hook)
            self.cc_hooks.append(handle)
        for hook in post_hooks:
            handle = self.register_forward_hook(hook)
            self.cc_hooks.append(handle)

    def construct(self, *args, **kwargs):
        return distributed_func.get(self.op_name_)(*args, **kwargs)

    def forward(self, *args, **kwargs):
        return distributed_func.get(self.op_name_)(*args, **kwargs)


class ApiRegistry:
    def __init__(self):
        self.distributed_attr_origin = {}
        self.distributed_attr_hooked = {}

    @staticmethod
    def store_ori_attr(ori_api_group, api_list, api_ori_attr):
        for api in api_list:
            if Const.SEP in api:
                sub_module_name, sub_op = api.rsplit(Const.SEP, 1)
                sub_module = getattr(ori_api_group, sub_module_name)
                api_ori_attr[api] = getattr(sub_module, sub_op)
            else:
                api_ori_attr[api] = getattr(ori_api_group, api)

    @staticmethod
    def set_api_attr(api_group, attr_dict):
        for cc_api_name, cc_api_entry_func in attr_dict.items():
            if Const.SEP in cc_api_name:
                sub_module_name, sub_op = cc_api_name.rsplit(Const.SEP, 1)
                sub_module = getattr(api_group, sub_module_name, None)
                if sub_module is not None:
                    setattr(sub_module, sub_op, cc_api_entry_func)
            else:
                setattr(api_group, cc_api_name, cc_api_entry_func)

    @staticmethod
    def redirect_wait():
        global ORIGIN_WAIT
        global PENDING_ASYNC_CC_BY_HANDLE
        if not ORIGIN_WAIT:
            return

        def wrapped_wait(work):
            def wrapped_wait(*args, **kwargs):
                ORIGIN_WAIT(*args, **kwargs)
                if args[0] in PENDING_ASYNC_CC_BY_HANDLE:
                    store_func = PENDING_ASYNC_CC_BY_HANDLE.pop(args[0])
                    store_func()

            return wrapped_wait

        CommHandle_.wait = wrapped_wait(CommHandle_)

    def redirect_api(self):
        self.set_api_attr(comm_func, self.distributed_attr_hooked)
        self.redirect_wait()

    def restore_api(self):
        if not ORIGIN_WAIT:
            return
        self.set_api_attr(comm_func, self.distributed_attr_origin)
        setattr(CommHandle_, 'wait', ORIGIN_WAIT)

    def reset_idx(self):
        for op_template in self.distributed_attr_hooked.values():
            if hasattr(op_template, "idx"):
                op_template.idx = 0

    def initialize_hook(self, pre_hooks, post_hooks):
        self.store_ori_attr(comm_func, get_distributed_ops(), self.distributed_attr_origin)
        cc_hooks = []
        for op_name in get_distributed_ops():
            self.distributed_attr_hooked[op_name] = DistributedOPTemplate(op_name, pre_hooks, post_hooks)
            cc_hooks.extend(self.distributed_attr_hooked[op_name].cc_hooks)
        return cc_hooks


def get_process_group(process_group):
    return (
        process_group
        if process_group
        else comm_func.HCCL_WORLD_GROUP
    )


def stack_filter(stack):
    for pattern in StackBlackList:
        if re.search(pattern, stack):
            return False
    return True


def get_callstack():
    callstack = []
    for (_, path, line, func, _, _) in inspect.stack():
        stack_line = f'{path}[{line}]'
        if stack_filter(stack_line):
            callstack.append(stack_line + '   ' + func)
    return callstack


@_no_grad()
def op_aggregate(op, tensorlist):
    if isinstance(tensorlist, Tensor):
        return tensorlist
    if not tensorlist:
        return Tensor(float('nan'))
    if op == 'min':
        return min(tensorlist)
    if op == 'max':
        return max(tensorlist)
    if op == 'norm':
        return sum(tensorlist)
    if op == 'zeros':
        return sum(tensorlist) / len(tensorlist)
    if op == 'nans':
        return sum(tensorlist)
    if op == 'mean':
        return sum(tensorlist) / len(tensorlist)
    return Tensor(float('nan'))


def update_data(old, new):
    for tag, op2tensor in new.items():
        if tag not in old:
            old[tag] = {}
        for op, tensor in op2tensor.items():
            if op not in old[tag]:
                old[tag][op] = [tensor]
            else:
                old[tag][op].append(tensor)
    return old


def is_target_line(codeline):
    stack = get_callstack()
    whole_stack = ';'.join(stack)
    if codeline == []:
        return True
    for pattern in codeline:
        if re.search(pattern, whole_stack):
            return True
    return False


@_no_grad()
def catch_data(params: CatchDataParams):
    tensor_args = {}
    for arg in params.args:
        if isinstance(arg, Tensor):
            key = get_summary_writer_tag_name(
                params.cc_name,
                f'{params.prefix}_{params.call_idx}_{len(tensor_args)}',
                RANK,
            )
            tensor_args[key] = arg
        elif isinstance(arg, list):
            if isinstance(arg[0], Tensor):
                stacked_arg = ops.stack(arg)
            elif isinstance(arg[0], comm_func.P2POp):
                stacked_arg = ops.stack([op.tensor for op in arg])
            key = get_summary_writer_tag_name(
                params.cc_name,
                f'{params.prefix}_{params.call_idx}_{len(tensor_args)}',
                RANK,
            )
            tensor_args[key] = stacked_arg

    new_data = get_metrics(params.ops_list, tensor_args, 1e-8)
    params.cc_context.data = update_data(params.cc_context.data, new_data)


def create_async_callback_func(params: CatchDataParams):
    def store_data():
        catch_data(params)

    return store_data


def create_hooks(context, monitor):
    def cc_log_hook(module, inputs):
        stack = ';'.join(get_callstack())
        monitor.cc_logged_stack[module.op_name_].add(stack)
        return

    def cc_pre_hook(module, inputs):
        if not is_target_line(monitor.cc_codeline):
            return
        call_idx = getattr(module, "idx", 0)
        catch_data(CatchDataParams(
            context[module.op_name_],
            module.op_name_,
            monitor.ops,
            inputs,
            MonitorConst.PREFIX_PRE,
            call_idx,
        ))
        return

    def cc_hook(module, inputs, out=None):
        if not is_target_line(monitor.cc_codeline):
            return out
        call_idx = getattr(module, "idx", 0)
        is_async_op = False
        if enable_communication:
            if isinstance(out, CommHandle_):
                PENDING_ASYNC_CC_BY_HANDLE[out] = create_async_callback_func(CatchDataParams(
                    context[module.op_name_],
                    module.op_name_,
                    monitor.ops,
                    inputs,
                    MonitorConst.PREFIX_POST,
                    call_idx,
                ))
                is_async_op = True
            elif isinstance(out, list):  # batch_isend_irecv
                for out_element in out:
                    if isinstance(out_element, comm_func.P2POp):
                        PENDING_ASYNC_CC_BY_HANDLE[out_element] = create_async_callback_func(CatchDataParams(
                            context[module.op_name_],
                            module.op_name_,
                            monitor.ops,
                            inputs,
                            MonitorConst.PREFIX_POST,
                            call_idx,
                        ))
                        is_async_op = True
            elif isinstance(out, tuple):
                if len(out) == 2 and isinstance(out[1], CommHandle_):
                    PENDING_ASYNC_CC_BY_HANDLE[out[1]] = create_async_callback_func(CatchDataParams(
                        context[module.op_name_],
                        module.op_name_,
                        monitor.ops,
                        inputs,
                        MonitorConst.PREFIX_POST,
                        call_idx,
                    ))
                    is_async_op = True
        if not is_async_op:
            out_list = [ii for ii in out if isinstance(ii, Tensor)]
            catch_data(CatchDataParams(
                context[module.op_name_],
                module.op_name_,
                monitor.ops,
                out_list,
                MonitorConst.PREFIX_POST,
                call_idx,
            ))
        module.idx = call_idx + 1
        return out

    global RANK
    pre_hooks = []
    hooks = []
    RANK = get_rank()
    if communication.GlobalComm.INITED and RANK not in monitor.module_rank_list and monitor.module_rank_list != []:
        return [pre_hooks, hooks]

    if monitor.cc_log_only:
        pre_hooks.append(cc_log_hook)
        return [pre_hooks, hooks]

    if monitor.cc_pre_hook:
        pre_hooks.append(cc_pre_hook)

    hooks.append(cc_hook)

    return [pre_hooks, hooks]


api_register = ApiRegistry()
